import argparse
import os

try:
    import rospy
    ROS_FOUND = True
except ImportError:
    ROS_FOUND = False


def parse_sim_args():
    # Arguments
    parser = argparse.ArgumentParser(description='Test node for the pepper RL path planning algorithm.')
    populate_sim_args(parser)
    # add external args
#     from xxx import parse_xxx_args
#     ARGS, unknown = parse_xxx_args(ignore_unknown=True, parser=parser)
    ARGS, unknown_args = parser.parse_known_args()

    # deal with unknown arguments
    # ROS appends some weird args, ignore those, but not the rest
    if unknown_args:
        if ROS_FOUND:
            non_ros_unknown_args = rospy.myargv(unknown_args)
        else:
            non_ros_unknown_args = unknown_args
        if non_ros_unknown_args:
            print("unknown arguments:")
            print(non_ros_unknown_args)
            parser.parse_args(args=["--help"])
            raise ValueError

    return ARGS


def populate_sim_args(parser):
    argscriptdir = os.path.dirname(os.path.realpath(__file__))
    parser.add_argument('--map-folder', type=str, default=os.path.join(argscriptdir, "../maps"),)
    parser.add_argument('--map-name', type=str, default="asl",)
    parser.add_argument('--map-downsampling-passes', type=int, default=0,)
    # planning args
    parser.add_argument('--n-trials', type=int, default=1000,)
    parser.add_argument('--max-depth', type=int, default=10,)
    single_action = parser.add_mutually_exclusive_group()
    single_action.add_argument('--only-intend', action='store_true',)
    single_action.add_argument('--only-say', action='store_true',)
    single_action.add_argument('--only-nudge', action='store_true',)
    # scenario args
    parser.add_argument(
        '--scenario-folder',
        type=str,
        default=os.path.join(argscriptdir, "../scenarios"),
    )
    parser.add_argument(
        '--scenario',
        type=str,
        default="aslquiet",
        help="""Which named scenario to load from the scenarios/ folder.
        The scenario defines initial agents and their state""",
    )
    # utilities args
    parser.add_argument('--debug', action='store_true',)
    parser.add_argument('--plotless', action='store_true',)



import pyIA.actions as actions

def agent_creator(agent_type):
    """ Creates an agent based on a type string """
#     print("creating agent of type {}".format(str(agent_type)))
    if "Robot" in str(agent_type):
        return Robot()
    if "Person" in str(agent_type):
        return Person()
    if "ORCAAgent" in str(agent_type):
        return ORCAAgent()
    else:
        raise ValueError("agent type {} does not exist".format(str(agent_type)))

## Agents
class Agent(object):
    """ Prototype for agent class """
    def type(self):
        return type(self)


class Robot(Agent):
    def __init__(self):
        self.radius = 0.3
        self.actions = [
                actions.Intend(),
                actions.Say(),
#                 actions.Crawl(),
                actions.Nudge(),
                ]

class Person(Agent):
    def __init__(self):
        self.radius = 0.3
        self.actions = [
#                 actions.Disturb(), # targeted
#                 actions.Intend(),
#                 actions.Loiter(),
                ]

# NIY or Deprecated ------------------------------
class MovableObstacle(Agent):
    def __init__(self):
        self.is_passive = True

class ORCAAgent(Agent):
    """ A cylindrical entity which moves around according to simple ORCA rules """
    def __init__(self):
        self.actions = [
                actions.Intend()
                ]
        self.radius = 0.3
# ------------------------------------------------

## inner states
class Innerstate(object):
    def __init__(self):
        self.patience = 1
        self.permissivity = 0.5
        self.perceptivity = 0.9
        self.goal = [0., 0.] # TODO: several goals with preference?
        self.undertaking = None





from __future__ import print_function
import copy
from CMap2D import CMap2D
from map2d import gridshow
from matplotlib import pyplot as plt
from matplotlib import patches as patches
import matplotlib
import numpy as np
import os
import pickle
import sys
import time

import pyIA.agents as agents
import pyIA.actions as actions
from pyIA.mdptree import Node

X_DIM = 2 # x, y

# scriptdir = os.path.dirname(os.path.realpath(__file__))
# SCENARIOS_FOLDER = os.path.join(scriptdir, "../scenarios")

class Worldstate(object):
    """ storage class for convenience
    Worlstate contains all possible information to define the state of the simulated world """
    def __init__(self):
        # agent state
        self._agentstate = {
            "agents": [], # list of agent prototypes
            "xystates": [], # [n_agents, 2]
            "uvstates": [],
            "innerstates": [],
        }
        # map state
        self.map = CMap2D()
        # others
        self.derived_state = DerivedWorldstate()
        # TODO: adding signals to the worldstate?
        # TODO: add undertakings to inner_state? (multi-step actions)

    def get_agents(self):
        return self._agentstate["agents"]
    def get_xystates(self):
        return self._agentstate["xystates"]
    def get_uvstates(self):
        return self._agentstate["uvstates"]
    def get_innerstates(self):
        return self._agentstate["innerstates"]

    def copy(self):
        """ non-trivial : agents, xy, uv, innerstates are copied but reference to the map and derived state is shared.
        Due to the fact that typically the map does not get modified in use-cases where copies of the worldstate are needed.
        Warning: Changing the map in a copy has side effects!"""
        copy_ = Worldstate()
        for key in self._agentstate:
            copy_._agentstate[key] = copy.deepcopy(self._agentstate[key])
        copy_.map = self.map
        copy_.derived_state = self.derived_state
        return copy_

    def remove_agent(self, agent_index):
        for key in self._agentstate:
            self._agentstate[key].pop(agent_index)

    def save_agents_to_file(self, filepath):
        with open(filepath,"wb") as f:
            pickle.dump(self._agentstate, f)
        print("Agentstate saved to {}".format(filepath))

    def restore_agents_from_file(self, filepath, silent=False):
        with open(filepath,"rb") as f:
            if sys.version_info.major >= 3:
                agentstate = pickle.load(f, encoding='latin1')
            else:
                agentstate = pickle.load(f)
        for key in agentstate:
            self._agentstate[key] = agentstate[key]
        # switch old agent classes for new agent classes
        for i in range(len(self._agentstate["agents"])):
            agent = self._agentstate["agents"][i]
            self._agentstate["agents"][i] = agents.agent_creator(agent.type())
        if not silent:
            print("Agentstate loaded from {}".format(filepath))



class DerivedWorldstate(object):
    """ convenience storage class for precomputed/cached fields,
    this class is for algorithmic optimization and thus isn't formally in the IA algo"""
    def __init__(self):
        self._map_sdf = None
        self._map_tsdf_2 = None
        self._map_as_closed_obst_vertices = None
        self._map_ij_arrays = None
        self._cache_map = None

    # CORE INTERFACE ------------------
    def copy(self):
        """ needed as otherwise, DerivedWorldstate shared between various states where goals differ
        might interfere with each other """
        copy_ = DerivedWorldstate()
        copy_._map_sdf = self._map_sdf
        copy_._map_tsdf_2 = self._map_tsdf_2
        copy_._map_as_closed_obst_vertices = self._map_as_closed_obst_vertices
        copy_._cache_map = self._cache_map
        return copy_

    def map_sdf(self, worldstate):
        if worldstate.map is not self._cache_map:
            self._on_set_map(worldstate)
        return self._map_sdf

    def map_tsdf_2(self, worldstate):
        if worldstate.map is not self._cache_map:
            self._on_set_map(worldstate)
        return self._map_tsdf_2

    def map_as_closed_obst_vertices(self, worldstate):
        if worldstate.map is not self._cache_map:
            self._on_set_map(worldstate)
        return self._map_as_closed_obst_vertices

    def map_ij_arrays(self, worldstate):
        if worldstate.map is not self._cache_map:
            self._on_set_map(worldstate)
        return self._map_ij_arrays

    def agent_precomputed_ctg(self, agent_index, worldstate):
        goal = worldstate.get_innerstates()[agent_index].goal
        goal_ij = worldstate.map.xy_to_ij(np.array([goal]))[0]
#         return worldstate.map.dijkstra(goal_ij, extra_costs=2-self.map_tsdf_2(worldstate), inv_value=-1, connectedness=16)
        return worldstate.map.dijkstra(
            goal_ij, extra_costs=1/(0.0000001 +self.map_sdf(worldstate)), inv_value=-1, connectedness=16)

    def density_map(self, worldstate, density_radius=2.):
        ijagents = [worldstate.map.xy_to_ij([xy])[0] for xy in worldstate.get_xystates()]
        ii, jj = self.map_ij_arrays(worldstate)
        density_map = np.zeros_like(ii)
        sqr_density_radius = (density_radius / worldstate.map.resolution())**2 
        for ai, aj in ijagents:
            density_map[((ii - ai)**2 + (jj - aj)**2) < sqr_density_radius] += 1
        return density_map

    def smooth_density_map(self, worldstate, density_radius=2.):
        map_ = worldstate.map
        ijagents = [worldstate.map.xy_to_ij([xy])[0] for xy in worldstate.get_xystates()]
        ii, jj = self.map_ij_arrays(worldstate)
        density_map = np.zeros_like(ii, dtype=np.float32)
        sqr_density_radius = (density_radius / map_.resolution())**2 
        for ai, aj in ijagents:
            mask = ((ii - ai)**2 + (jj - aj)**2) < sqr_density_radius
            density_map[mask] += 1 - np.sqrt((ii - ai)**2 + (jj - aj)**2)[mask]/ np.sqrt(sqr_density_radius)
        return density_map

    def inv_density_map(self, worldstate, density_radius=2.):
        map_ = worldstate.map
        ijagents = [worldstate.map.xy_to_ij([xy])[0] for xy in worldstate.get_xystates()]
        ii, jj = self.map_ij_arrays(worldstate)
        density_map = np.zeros_like(ii, dtype=np.float32)
        sqr_density_radius = (density_radius / map_.resolution())**2 
        for ai, aj in ijagents:
            mask = ((ii - ai)**2 + (jj - aj)**2) < sqr_density_radius
            density_map[mask] += 1. / (np.sqrt((ii - ai)**2 + (jj - aj)**2)[mask]*4/np.sqrt(sqr_density_radius) + 0.000001)
        return density_map


    def visibility_map_ij(self, agent_index, worldstate):
        ij = worldstate.map.xy_to_ij([worldstate.get_xystates()[agent_index]])[0]
        visibility = worldstate.map.visibility_map_ij(ij)
        return visibility

    # INTERNAL -------------------------
    def _on_set_map(self, worldstate):
        """ Calculates and caches derived state for obstacle map """
        self._cache_map = worldstate.map
        self._map_sdf = worldstate.map.as_sdf()
        self._map_tsdf_2 = worldstate.map.as_tsdf(2)
        self._map_as_closed_obst_vertices = worldstate.map.as_closed_obst_vertices()
        from pyniel.numpy_tools import indexing
        ij = indexing.as_idx_array(worldstate.map.occupancy(), axis='all')
        self._map_ij_arrays = (ij[...,0], ij[...,1])

    def check_is_up_to_date(self):
        """ compares checksum for latest value updates with current checksums """
        raise NotImplementedError


class Environment(object):
    def __init__(self, args, silent=False):
        self.worldstate = Worldstate()
        map_ = CMap2D(args.map_folder, args.map_name, silent=silent)
        for _ in range(args.map_downsampling_passes):
            map_ = map_.as_coarse_map2d()
        self.worldstate.map = map_
        self.args = args

    # CORE INTERFACE -------------------
    def step(self, actions):
        # high level (ia action consequences) or low level (execute planners)
        pass

    def get_worldstate(self):
        return self.worldstate

    def get_sensor_readings(self, agent_index):
        # ???
        pass

    # INTERNAL INTERFACE ---------------
    def fullstate(self):
        relstate = self.relstate(self.worldstate)
        return (self.worldstate, relstate)

    def relstate(self, worldstate): # DEPRECATED
        n_agents = len(worldstate.get_agents())
        xystates = np.array(worldstate.get_xystates())
        deltas = xystates[:, None, :] - xystates[None, :, :]
        drelstates = np.linalg.norm(deltas, axis=-1)
        threlstates = np.arctan2(deltas[1], deltas[0])
        # TODO: rel uv
        return (drelstates, threlstates)

    def policies(self, fullstate):
        """ runs a policy simulation on each agent,
        returns each agents chosen X_desired, and action """
        worldstate, relstate = fullstate
        for i, _ in enumerate(worldstate.get_agents()):
            agent_i = worldstate.get_agents()[i]
            innerstate_i = worldstate.get_innerstates()[i]
            xystate_i = worldstate.get_xystates()[i]
            relstate_i = relstate[0][i]
            # build mdp tree
            node0 = Node(worldstate, [], [], 0, [])
            mdpt = [node0]
            # iterate through potential actions
#             for a in agent_i.actions:
#                 outcome_distribution = a.old_predict_outcome(i, worldstate)
#                 for outcome, prob in outcome_distribution:
#                     pass


        A = []
        Xd = []
        return (A, Xd) # actions, desired X for each agent

    def enact_policies(self, A, Xd, worldstate):
        """ Once policies are chosen, they can be enacted to advance to a new state.
        Two alternatives, execute_planners, and predict_planners. """
        return predict_planners(A, Xd, worldstate)

    def execute_planners(self, A, Xd, worldstate):
        """ Runs a low level simulation of the agents and their action planners interacting,
        outputs resulting positions (Xr), and whether the planner succeeded  """
        return (Xr, success)

    def predict_planners(self, A, Xd, worldstate):
        """ Tries to directly predict positions (Xr) and success of the planners by 
        using simple interaction models of the planners """
        return (Xr, success)

    # SAMPLING -------------------------
    def x_samples(self, agent, innerstate, xystate, relstate, nudge_stddev=0.01):
        import hexgrid
        """ Generate a graph of X samples starting from an agent's X state """
        # neighbor nodes in hexagonal grid
        neighbors = hexgrid.neighbors(coordinate_system="xy")
        neighbors = np.concatenate((neighbors, hexgrid.neighbors(coordinate_system="xy", hex_orientation="pointy")), axis=0)
        nudge = np.random.normal(0, nudge_stddev, size=neighbors.shape)
        x_samples = xystate + neighbors + nudge
        return x_samples

    # VISUALIZATIONS -------------------
    def plot_map(self):
        gridshow(self.worldstate.map.occupancy())

    def plot_map_contours_ij(self, *args, **kwargs):
        if not args:
            args = ('-,',)
        contours = [self.worldstate.map.xy_to_ij(c) for c in self.worldstate.derived_state.map_as_closed_obst_vertices(self.worldstate)]
        for c in contours:
            # add the first vertice at the end to close the plotted contour
            cplus = np.concatenate((c, c[:1, :]), axis=0)
            plt.plot(cplus[:,0], cplus[:,1], *args, **kwargs)

    def plot_map_contours_xy(self, *args, **kwargs):
        if not args:
            args = ('-,',)
        contours = self.worldstate.derived_state.map_as_closed_obst_vertices(self.worldstate)
        for c in contours:
            # add the first vertice at the end to close the plotted contour
            cplus = np.concatenate((c, c[:1, :]), axis=0)
            plt.plot(cplus[:,0], cplus[:,1], *args, **kwargs)

    def plot_agents_xy(self, figure=None):
        if figure is None:
            figure = plt.gcf()
        patches = []
        worldstate = self.worldstate
        for i, _ in enumerate(worldstate.get_agents()):
            patch = self.get_viz_agent_patch(i)
            figure.gca().add_artist(patch)
            patches.append(patch)
        return patches

    def get_viz_agent_patch(self, agent_index):
        worldstate = self.worldstate
        agent = worldstate.get_agents()[agent_index]
        xy = np.array(worldstate.get_xystates()[agent_index])
        c = self.get_viz_agent_color(agent_index)
        r = agent.radius
        patchcreator = self.get_viz_agent_patchcreator(agent)
        patch = patchcreator((xy[0], xy[1]), r, facecolor=c, edgecolor="black")
        return patch

    def get_viz_agent_color(self, agent_index):
        worldstate = self.worldstate
        GrYlRd = matplotlib.colors.LinearSegmentedColormap.from_list("", ["tomato","yellow","springgreen"])
        c = GrYlRd(worldstate.get_innerstates()[agent_index].permissivity)
        return c

    def get_viz_agent_patchcreator(self, agent):
        if agent.type() == agents.Robot:
            def PatchCreator(xy, r, **kwargs):
                return patches.RegularPolygon(xy, 6, r, **kwargs)
        elif agent.type() == agents.ORCAAgent:
            def PatchCreator(xy, r, **kwargs):
                return patches.RegularPolygon(xy, 4, r, **kwargs)
        else:
            def PatchCreator(xy, r, **kwargs):
                return plt.Circle(xy, r, **kwargs)
        return PatchCreator

    def plot_agents_ij(self):
        worldstate = self.worldstate
        for i, a in enumerate(worldstate.get_agents()):
            xy = np.array(worldstate.get_xystates()[i])
            ij = self.worldstate.map.xy_to_ij([xy])[0]
            GrYlRd = matplotlib.colors.LinearSegmentedColormap.from_list("", ["tomato","yellow","springgreen"])
            c = GrYlRd(worldstate.get_innerstates()[i].permissivity)
            r = a.radius / self.worldstate.map.resolution()
            if a.type() == agents.Robot:
                circle = patches.RegularPolygon((ij[0], ij[1]), 6, r, color=c)
                plt.gcf().gca().add_artist(circle)
#                 marker = "^"
#                 plt.scatter(ij[:1], ij[1:], marker=marker, c=[c])
            else:
                circle = plt.Circle((ij[0], ij[1]), r, color=c)
                plt.gcf().gca().add_artist(circle)

    def plot_goals(self):
        ij_goals = self.worldstate.map.xy_to_ij(np.array([innerstate.goal for innerstate in self.worldstate.get_innerstates()]))
        plt.scatter(ij_goals[:,0], ij_goals[:,1], marker="D", color='y')



    def plot_samples(self):
        pass


    # LOGISTICS --------------------------------------- 
    def populate(self, silent=False):
        """ add agents according to some scenario """
        if self.args.scenario == "":
            pass
        else:
            filepath = os.path.join(self.args.scenario_folder, self.args.scenario+".pickle")
            self.worldstate.restore_agents_from_file(filepath, silent=silent)




from __future__ import print_function
import numpy as np
from pyniel.numpy_tools import indexing
import warnings

from cia import CState, FixedState
from cia import kStateFeatures

# State estimation
PERMISSIVITY_RADIUS = 1.
PERCEPTIVITY_RADIUS = 1.
CROWDEDNESS_RADIUS = 1.

MEMORY_FORGET_RATE = 1.2  # stddev multiplicator, /sec
BIG_SIGMASQ = 1000.  # stddev larger than this value is treated like infinite for numerical purp.
TINY_SIGMASQ = 0.00001  # stddev smaller than this is ... 0 for numerical purp.
CHECK_NUMERICS = True

# initial P(S)
CROWDEDNESS_UNINFORMED_PRIOR = 0.1
PERMISSIVITY_UNINFORMED_PRIOR = 0.99
PERCEPTIVITY_UNINFORMED_PRIOR = 0.99
CWD_UNCERTAINTY_UNINFORMED_PRIOR = BIG_SIGMASQ
PRM_UNCERTAINTY_UNINFORMED_PRIOR = BIG_SIGMASQ
PER_UNCERTAINTY_UNINFORMED_PRIOR = BIG_SIGMASQ
# P(Z|S) parameters mu, sigma^2 for empty observations
NOHUMAN_CROWDEDNESS_SENSOR_PR = 0.
NOHUMAN_PERMISSIVITY_SENSOR_PR = 0.99
NOHUMAN_PERCEPTIVITY_SENSOR_PR = 0.99
# TODO: issue with occluded humans (for now boosted uncertainties on absence of detection)
NOHUMAN_CWD_UNCERTAINTY_SENSOR_PR = 2.
NOHUMAN_PRM_UNCERTAINTY_SENSOR_PR = 2.
NOHUMAN_PER_UNCERTAINTY_SENSOR_PR = 2.
# P(Z|S) parameters mu, sigma^2 for human detections
HUMAN_CROWDEDNESS_SENSOR_PR = 1.
HUMAN_PERMISSIVITY_SENSOR_PR = 0.5
HUMAN_PERCEPTIVITY_SENSOR_PR = 0.8
# few false positives with YOLO (some smearing due to RGB-D)
HUMAN_CWD_UNCERTAINTY_SENSOR_PR = 0.05
HUMAN_PRM_UNCERTAINTY_SENSOR_PR = 2.  # observing a person tells me little about their PRM
HUMAN_PER_UNCERTAINTY_SENSOR_PR = 0.1

# sensor properties
HALF_FOV = np.pi / 4.  # 45 degrees ( total 90 deg realsense fov )

# CORE INTERFACE -----------
def true_state_from_sim_worldstate(agent_index, worldstate):
    raise NotImplementedError

def state_estimate_from_sim_worldstate(agent_index, worldstate,
                                       state_memory=None, time_elapsed=None,
                                       reuse_state_cache=None):
    pos_xy = worldstate.get_xystates()[agent_index]  # + Noise?
    pos_ij = worldstate.map.xy_to_floatij([pos_xy])[0]
    # fixed state
    if reuse_state_cache is not None:
        fixed_state = reuse_state_cache
    else:
        fixed_state = FixedState(worldstate.map)
    # prior
    state = CState(
        radius=worldstate.get_agents()[agent_index].radius,
        pos=np.array(pos_xy, dtype=np.float32),
        pos_ij=np.array(pos_ij, dtype=np.float32),
        mapwidth=worldstate.map.occupancy().shape[0],
        mapheight=worldstate.map.occupancy().shape[1],
        goal=np.array(worldstate.get_innerstates()[agent_index].goal, dtype=np.float32),
    )
    if state_memory is not None:
        merge_state_memory_into_state(state, state_memory, time_elapsed)
    else:
        uninformed_prior_into_state(state, fixed_state)
    # apply sensor update to prior
    grid_feature_estimate_from_sim_worldstate(
        agent_index, worldstate, state)
    return state, fixed_state

def state_estimate_from_tracked_persons(tracked_persons, pose2d_tracksframe_in_refmap,
                                        robot_pos, robot_radius, goal_xy, map_,
                                        state_memory=None, time_elapsed=None,
                                        reuse_state_cache=None):
    """
    pos_xy is robot position in map
    """
    pos_xy = robot_pos[:2]
    robot_heading = robot_pos[2]
    pos_ij = map_.xy_to_floatij([pos_xy])[0]
    # fixed state
    if reuse_state_cache is not None:
        fixed_state = reuse_state_cache
    else:
        fixed_state = FixedState(map_)
    # prior
    state = CState(
        radius=robot_radius,
        pos=np.array(pos_xy, dtype=np.float32),
        pos_ij=np.array(pos_ij, dtype=np.float32),
        mapwidth=map_.occupancy().shape[0],
        mapheight=map_.occupancy().shape[1],
        goal=np.array(goal_xy, dtype=np.float32),
    )
    if state_memory is not None:
        merge_state_memory_into_state(state, state_memory, time_elapsed)
    else:
        uninformed_prior_into_state(state, fixed_state)
    # apply sensor update to prior
    grid_feature_estimate_from_sensor_data(
        tracked_persons, pose2d_tracksframe_in_refmap, robot_heading, state, fixed_state)
    return state, fixed_state

# TODO Add lidar for free space confirmation

# INTERNAL -----------------
def tracked_persons_from_sim_worldstate(robot_index, worldstate):
    from frame_msgs.msg import TrackedPersons, TrackedPerson
    tracked_persons = TrackedPersons()
    n_agents = len(worldstate.get_agents())
    robot_ij = worldstate.map.xy_to_ij([worldstate.get_xystates()[robot_index]])[0]
    robot_heading = agent_heading_from_sim_worldstate(robot_index, worldstate)
    # visibility_map  # TODO: use lidar free space instead!
    fov = [robot_heading - HALF_FOV, robot_heading + HALF_FOV]
    visibility = worldstate.map.visibility_map_ij(robot_ij, fov=fov)
    for k in range(n_agents):
        if k == robot_index:
            continue
        axy = worldstate.get_xystates()[k]
        auv = worldstate.get_uvstates()[k]
        ai, aj = worldstate.map.xy_to_ij([axy])[0]
        agent_heading = agent_heading_from_sim_worldstate(k, worldstate)
        if visibility[ai, aj] < 0:
            continue
        tp = TrackedPerson()
        from tf.transformations import quaternion_from_euler
        quaternion = quaternion_from_euler(0, 0, agent_heading)
        tp.pose.pose.orientation.x = quaternion[0]
        tp.pose.pose.orientation.y = quaternion[1]
        tp.pose.pose.orientation.z = quaternion[2]
        tp.pose.pose.orientation.w = quaternion[3]
        tp.pose.pose.position.x = axy[0]
        tp.pose.pose.position.y = axy[1]
        tp.twist.twist.linear.x = auv[0]
        tp.twist.twist.linear.y = auv[1]
        tracked_persons.tracks.append(tp)
    pose2d_tracksframe_in_refmap = np.array([0, 0, 0])
    return tracked_persons, pose2d_tracksframe_in_refmap

def agent_heading_from_sim_worldstate(agent_index, worldstate):
    uv = worldstate.get_uvstates()[agent_index]
    if np.linalg.norm(uv) == 0:
        pos_xy = worldstate.get_xystates()[agent_index]
        goal_xy = worldstate.get_innerstates()[agent_index].goal
        heading = np.arctan2(goal_xy[1] - pos_xy[1], goal_xy[0] - pos_xy[0])
    else:
        heading = np.arctan2(uv[1], uv[0])
    return heading

def grid_feature_estimate_from_sim_worldstate(agent_index, worldstate, state):
    tracked_persons, pose2d_tracksframe_in_refmap = \
        tracked_persons_from_sim_worldstate(agent_index, worldstate)
    agent_heading = agent_heading_from_sim_worldstate(agent_index, worldstate)
    fixed_state = FixedState(worldstate.map)
    grid_feature_estimate_from_sensor_data(tracked_persons, pose2d_tracksframe_in_refmap,
                                           agent_heading, state, fixed_state)

def grid_feature_estimate_from_sensor_data(tracked_persons, pose2d_tracksframe_in_refmap,
                                           robot_heading, state, fixed_state):
    # same for all features
    n_agents = len(tracked_persons.tracks)
    # visibility_map  # TODO: use lidar free space instead!
    map_ = fixed_state.map
    ij = indexing.as_idx_array(map_.occupancy(), axis='all')
    ii = ij[..., 0]
    jj = ij[..., 1]
    fov = [robot_heading - HALF_FOV, robot_heading + HALF_FOV]
    visibility = fixed_state.map.visibility_map_ij(state.get_pos_ij(), fov=fov)
    for feature_name in kStateFeatures:
        feature_index = kStateFeatures[feature_name]
        # start by assuming no humans detected, then add detections one by one.
        # TODO: increase nohuman uncertainty with distance?
        if feature_name == "crowdedness":
            feature_map = NOHUMAN_CROWDEDNESS_SENSOR_PR * np.zeros_like(ii, dtype=np.float32)
            feature_uncertainty_map = NOHUMAN_CWD_UNCERTAINTY_SENSOR_PR * np.ones_like(ii, dtype=np.float32)
        elif feature_name == "permissivity":
            feature_map = NOHUMAN_PERMISSIVITY_SENSOR_PR * np.ones_like(ii, dtype=np.float32)
            feature_uncertainty_map = NOHUMAN_PRM_UNCERTAINTY_SENSOR_PR * np.ones_like(ii, dtype=np.float32)
        elif feature_name == "perceptivity":
            feature_map = NOHUMAN_PERCEPTIVITY_SENSOR_PR * np.ones_like(ii, dtype=np.float32)
            feature_uncertainty_map = NOHUMAN_PER_UNCERTAINTY_SENSOR_PR * np.ones_like(ii, dtype=np.float32)
        for k in range(n_agents):
            tp = tracked_persons.tracks[k]
            from tf.transformations import euler_from_quaternion
            import pose2d
            quaternion = (
                tp.pose.pose.orientation.x,
                tp.pose.pose.orientation.y,
                tp.pose.pose.orientation.z,
                tp.pose.pose.orientation.w,
            )
            tp_pos_in_tracksframe = np.array([
                tp.pose.pose.position.x,
                tp.pose.pose.position.y,
                euler_from_quaternion(quaternion)[2],
            ])
            tp_pos_in_refmap = pose2d.apply_tf_to_pose(
                tp_pos_in_tracksframe, pose2d_tracksframe_in_refmap)
            ai, aj = map_.xy_to_ij([tp_pos_in_refmap[:2]])[0]
            # compute P(Z|S)
            if feature_name == "crowdedness":
                density_radius = CROWDEDNESS_RADIUS
            elif feature_name == "permissivity":
                density_radius = PERMISSIVITY_RADIUS
            elif feature_name == "perceptivity":
                density_radius = PERCEPTIVITY_RADIUS
            else:
                raise NotImplementedError("state estimate for {} not implemented.".format(
                    feature_name))
            sqr_density_radius_ij = (density_radius / map_.resolution())**2
            sqr_distance = ((ii - ai)**2 + (jj - aj)**2)
            mask = sqr_distance < sqr_density_radius_ij
            if feature_name == "crowdedness":
#                 influence_radius = 1.  # bigger means "larger" people
#                 feature_map[mask] += np.clip(1. / (0.0001 + sqr_distance[mask] * influence_radius),
#                                              0., 1.)
                feature_map[mask] += HUMAN_CROWDEDNESS_SENSOR_PR  # additive
                feature_uncertainty_map[mask] = HUMAN_CWD_UNCERTAINTY_SENSOR_PR
            elif feature_name == "permissivity":
                feature_map[mask] = HUMAN_PERMISSIVITY_SENSOR_PR
                feature_uncertainty_map[mask] = HUMAN_PRM_UNCERTAINTY_SENSOR_PR
            elif feature_name == "perceptivity":
                feature_map[mask] = HUMAN_PERCEPTIVITY_SENSOR_PR
                feature_uncertainty_map[mask] = HUMAN_PER_UNCERTAINTY_SENSOR_PR
        # what you can't see...
        feature_uncertainty_map[visibility < 0] = np.inf  # no sensor update outside of sensor fov
        # apply observation normal pdf using bayesian posterior  P(S|Z) ~= P(S) * P(Z|S)
        # product of two gaussian pdfs
        mu1 = np.array(state.grid_features_values()[feature_index, :, :])
        mu2 = feature_map
        sigmasq1 = np.array(state.grid_features_uncertainties()[feature_index, :, :])
        sigmasq2 = feature_uncertainty_map
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', r'invalid value encountered')
            warnings.filterwarnings('ignore', r'divide by zero')
            new_mu = (mu1 * sigmasq2 + mu2 * sigmasq1) / (sigmasq2 + sigmasq1)
            new_sigmasq = 1. / (1. / sigmasq1 + 1. / sigmasq2)
        # check numerics
        if CHECK_NUMERICS:
            is_both_small = np.logical_and(sigmasq1 < TINY_SIGMASQ, sigmasq2 < TINY_SIGMASQ)
            is_1_big = sigmasq1 > BIG_SIGMASQ
            is_2_big = sigmasq2 > BIG_SIGMASQ
            is_both_big = np.logical_and(sigmasq1 > BIG_SIGMASQ, sigmasq2 > BIG_SIGMASQ)
            new_mu[is_both_small] = ((mu1 + mu2) / 2.)[is_both_small]
            new_mu[is_1_big] = mu2[is_1_big]
            new_mu[is_2_big] = mu1[is_2_big]
            new_mu[is_both_big] = ((mu1 + mu2) / 2.)[is_both_big]
            new_sigmasq[is_both_small] = TINY_SIGMASQ
            new_sigmasq[is_1_big] = sigmasq2[is_1_big]
            new_sigmasq[is_2_big] = sigmasq1[is_2_big]
            new_mu[is_both_big] = BIG_SIGMASQ
            new_sigmasq[np.isnan(new_sigmasq)] = np.inf  # this shouldn't happen... so it will
        # assign
        state.grid_features_values()[feature_index, :, :] = new_mu
        state.grid_features_uncertainties()[feature_index, :, :] = new_sigmasq

def merge_state_memory_into_state(state, state_memory, time_elapsed):
    """ should be split into two functions:
    1. copies memory prior into state
    2. adds forgetfulness update to memory prior
    """
    for feature_name in kStateFeatures:
        feature = kStateFeatures[feature_name]
        # copy prior into state
        state.grid_features_values()[feature] = \
            1. * state_memory.grid_features_values()[feature]
        state.grid_features_uncertainties()[feature] = \
            1. * state_memory.grid_features_uncertainties()[feature]
        # knowledge that state features are transient -> increase uncertainty over time
        forgetfulness = np.power(MEMORY_FORGET_RATE, time_elapsed)
        state.grid_features_uncertainties()[feature] = \
            forgetfulness * state.grid_features_uncertainties()[feature]
        if np.any(np.isnan(state.grid_features_uncertainties()[feature])):
            print("IA SE: nan values in memory update")
            state.grid_features_uncertainties()[feature][
                np.isnan(state.grid_features_uncertainties()[feature])
            ] = np.inf
            state.grid_features_uncertainties()[feature][
                state.grid_features_uncertainties()[feature] > BIG_SIGMASQ
            ] = BIG_SIGMASQ

def uninformed_prior_into_state(state, fixed_state):
    """ The absolutely uninformed prior.
    How you expect the state to be in a place you've never seen """
    map_ = fixed_state.map
    occ = map_.occupancy()
    for feature_name in kStateFeatures:
        feature_index = kStateFeatures[feature_name]
        # feature_map
        if feature_name == "crowdedness":
            feature_map = CROWDEDNESS_UNINFORMED_PRIOR * np.ones_like(occ, dtype=np.float32)
            feature_uncertainty_map = CWD_UNCERTAINTY_UNINFORMED_PRIOR * np.ones_like(occ, dtype=np.float32)
        elif feature_name == "permissivity":
            feature_map = PERMISSIVITY_UNINFORMED_PRIOR * np.ones_like(occ, dtype=np.float32)
            feature_uncertainty_map = PRM_UNCERTAINTY_UNINFORMED_PRIOR * np.ones_like(occ, dtype=np.float32)
        elif feature_name == "perceptivity":
            feature_map = PERCEPTIVITY_UNINFORMED_PRIOR * np.ones_like(occ, dtype=np.float32)
            feature_uncertainty_map = PER_UNCERTAINTY_UNINFORMED_PRIOR * np.ones_like(occ, dtype=np.float32)
        state.grid_features_values()[feature_index, :, :] = feature_map
        state.grid_features_uncertainties()[feature_index, :, :] = feature_uncertainty_map



from __future__ import print_function
import copy
from heapq import heappush, heappop
from CMap2D import gridshow
import matplotlib.pyplot as plt
import numpy as np
from pyniel.pyplot_tools.realtime import plot_closeall_button
import warnings

import pyIA.paths as paths
from cia import segment_endindex_at_pathpos_i
from cia import path_state_from_state_and_path
from cia import sultans_wife
from cia import bool1d_from_segments
from cia import CTaskState
from cia import kStateFeatures

# Planner
ACTION_CMAP = plt.cm.rainbow
ACTION_LENGTH_LIMIT_IN_TASK_TREE = 6.  # meters

# STATE_FEATURES = ... # list of state features has been moved to cia.pyx
DYNAMIC_STATE = False  # the whole state is dynamic (useful only if paths can change within MCTS)
DYNAMIC_PATH_STATE = True  # only the path state is dynamic
STORE_ESTIMATE_DISTRIBUTIONS = True
ALLOW_CONSECUTIVE_FAILURE = True



class Branch(object):
    def __init__(self, target, action, meta={
            "success": None,
            "cost": None,
            "prob": None,
    }):
        self.target = target  # id of node at the branch end
        self.action = action
        self.meta = meta


class BranchBundle(object):
    def __init__(self, meta={
            "tasktargetpathidx": None,
            "taskaction": None,
            "taskfullpath": None,
    }):
        self.branches = []  # [Branch(), ...]
        self.meta = meta


class TTNode(object):
    def __init__(self):
        self.childbranches = set()


class STNode(object):
    def __init__(self,
                 next_id,
                 depth,
                 startidx,
                 state,
                 path_state,
                 path_variant,
                 task_tree,
                 possible_actions,
                 cumulative_cost=None,
                 cumulative_prob=None,
                 ):
        # metadata
        self.id = next_id[0]  # "pass by reference" hack
        next_id[0] += 1
        self.depth = depth
        self.is_open = True
        # data
        self.startidx = startidx  # along path
        # edges
        self.childbranchbundles = []
        # indirect data
        self.cumulative_cost = cumulative_cost
        self.cumulative_prob = cumulative_prob
        # meta
        self.mce_info = None  # monte carlo estimate info can be stored here as NodeMCEstimateInfo()
        # necessary for lazy expansion (should only be set if node is open and yet to be updated)
        self.task_state_update = None
        # lazy expansion data (not know until the node is closed)
        self.is_terminal = False
        self.state = state
        self.path_state = path_state
        self.path_variant = path_variant
        self.task_tree = task_tree
        self.possible_actions = possible_actions

    def copy(self):
        # only the childbranchbundles are deepcopied! nodes are generally not copied.
        copy_ = STNode(
            [self.id],
            self.depth,
            self.startidx,
            self.state.copy(),
            self.path_state,  # TODO copy()
            self.path_variant,
            self.task_tree,
            self.possible_actions,
        )
        copy_.is_open = self.is_open
        copy_.childbranchbundles = copy.deepcopy(self.childbranchbundles)
        copy_.cumulative_cost = self.cumulative_cost
        copy_.cumulative_prob = self.cumulative_prob
        copy_.mce_info = copy.deepcopy(self.mce_info)
        copy_.task_state_update = self.task_state_update
        copy_.is_terminal = self.is_terminal
        return copy_


class MCBranchEstimate(object):
    def __init__(self):
        self.n_goal_reached = 0  # times goal was reached when this node/branch was selected
        self.n_total = 0  # times this node/branch was selected
        self.sum_cost_to_goal = 0  # total final cost of all trials where this node/branch was selected
        # optional, store distribution of outcome costs
        self.dist = []

    def expected_final_cost(self):
        if self.n_goal_reached == 0:
            return np.nan
        return 1. * self.sum_cost_to_goal / self.n_goal_reached

    def prob_reach_goal(self):
        return 1. * self.n_goal_reached / self.n_total


class NodeMCEstimateInfo(object):
    def __init__(self, option_mc_estimates):
        self.option_mc_estimates = option_mc_estimates
        self.endnode_estimate = MCBranchEstimate()  # estimate for when MC algo ends on this node

    def node_estimate(self):
        mc_estimate = MCBranchEstimate()
        # sum branch estimates
        for est in self.option_mc_estimates:
            mc_estimate.n_goal_reached += est.n_goal_reached
            mc_estimate.n_total += est.n_total
            mc_estimate.sum_cost_to_goal += est.sum_cost_to_goal
            mc_estimate.dist.extend(est.dist)
        # add terminal estimates for self
        mc_estimate.n_goal_reached += self.endnode_estimate.n_goal_reached
        mc_estimate.n_total += self.endnode_estimate.n_total
        mc_estimate.sum_cost_to_goal += self.endnode_estimate.sum_cost_to_goal
        mc_estimate.dist.extend(self.endnode_estimate.dist)
        return mc_estimate


class StochasticTree(object):
    def __init__(self):
        self.nodes = {}
        # meta
        self.root_nodes = []
        self.next_id = [0]
        self.fixed_state = None


# CORE interface -----------
def path_options(state, fixed_state, possible_path_variants=None, debug=False):
    if possible_path_variants is None:
        possible_path_variants = paths.PATH_VARIANTS
    paths_xy = []
    path_variants = []
    for variant in possible_path_variants:
        path_xy = fixed_state.derived_state.path_to_goal(state, fixed_state, variant)
        if path_xy is None:
            continue
        if paths.is_path_in_list(path_xy, paths_xy):
            continue
        paths_xy.append(path_xy)
        path_variants.append(variant)
    if debug:
        plt.show()
    return [np.array(path, dtype=np.float32) for path in paths_xy], path_variants

def task_tree_from_path_state(path_state, possible_actions,
                              task_length_limit=None, constrain_to_depth_1=False):
    path_len = path_state.get_path_len()
    action_segment_options = []
    # check action preconditions
    for action in possible_actions:
        proto_segments, segment_constraints = action.check_preconditions_on_path(path_state)
        proto_endindices = segment_endindex_at_pathpos_i(proto_segments, path_state)
        action_segment_options.append(
            (action, proto_segments, proto_endindices, segment_constraints)
        )
    # trigger points
    trigger_points = {-1}  # -1 is a convenience to point to the segment end
    for option in action_segment_options:
        _, proto_segments, _, _ = option
        for seg in proto_segments:
            if seg[0] != 0:
                trigger_points.add(seg[0])
    # build tree by combining valid action proto-segments
    firstidx = path_state.pos_along_path
    open_nodes = {firstidx}  # set {nodeindex}
    closed_nodes = {path_len - 1: TTNode()}  # dict {nodeindex: TTNode}
    while True:
        if not open_nodes:
            break
        actionstartidx = open_nodes.pop()
        node = TTNode()
        for option in action_segment_options:
            action, _, proto_endindices, segment_constraints = option
            is_length_limit_trigger_point_added = False  # avoid putting the same trigger point twice
            # check if action can be started at this point in path
            if proto_endindices[actionstartidx] == -1:
                continue
            # multiple possibilities for ending: end at segment end or trigger point
            for t in trigger_points:
                if t == -1:
                    actionendidx = proto_endindices[actionstartidx]
                # only allow triggerpoints in the future
                elif t <= actionstartidx:
                    continue
                # don't allow triggerpoints outside of action preconditions
                elif t - 1 > proto_endindices[actionstartidx]:
                    continue
                else:
                    # set action to end right before the trigger point
                    # this way the next action will start AT the trigger point
                    actionendidx = t - 1
                # add hacky length constraint
                if task_length_limit is not None:
                    if actionendidx > actionstartidx + task_length_limit:
                        if is_length_limit_trigger_point_added:
                            continue
                        is_length_limit_trigger_point_added = True
                        actionendidx = actionstartidx + task_length_limit
                # check if constraints are held by taking this action
                if not segment_constraints((actionstartidx, actionendidx), path_state):
                    continue
                # add action to the tree
                resultidx = actionendidx + 1
                if resultidx >= path_len:
                    resultidx = path_len - 1
                node.childbranches.add(Branch(resultidx, action))
                if resultidx not in closed_nodes:
                    open_nodes.add(resultidx)
        closed_nodes[actionstartidx] = node
        if constrain_to_depth_1:
            for node_startidx in open_nodes:
                closed_nodes[node_startidx] = TTNode()
            break
    task_tree = closed_nodes
    return task_tree

def plan(state, fixed_state, possible_actions,
         n_trials=1000, max_depth=10,
         BYPRODUCTS=None, DEBUG=False, INTERRUPT=None, DEBUG_IF_FAIL=True):
    byproducts = {}
    if BYPRODUCTS is not None:
        byproducts = BYPRODUCTS
    paths_xy, path_variants = path_options(state, fixed_state)
    byproducts["paths_xy"] = paths_xy
    byproducts["path_variants"] = path_variants
    path_states = [path_state_from_state_and_path(state, path_xy, fixed_state.map.xy_to_floatij(path_xy))
                   for path_xy in paths_xy]
    byproducts["path_states"] = path_states
    task_length_limit = None
    if ACTION_LENGTH_LIMIT_IN_TASK_TREE is not None:
        task_length_limit = int(ACTION_LENGTH_LIMIT_IN_TASK_TREE / fixed_state.map.resolution())
    task_trees = [task_tree_from_path_state(path_state, possible_actions, task_length_limit)
                  for path_state in path_states]
    byproducts["task_trees"] = task_trees
    stochastic_tree = initialize_stochastic_tree(
        state, fixed_state, path_states, path_variants, task_trees, possible_actions, DEBUG=DEBUG)
    byproducts["stochastic_tree"] = stochastic_tree
    # Detect Error
    if not stochastic_tree.root_nodes:
        if DEBUG_IF_FAIL:
            print("Error: no possibilities found.")
            visualize_state_features(state, fixed_state, hide_uncertain=False)
            visualize_traversabilities(state, fixed_state, paths.PATH_VARIANTS)
            visualize_dijkstra_fields(state, fixed_state, paths.PATH_VARIANTS)
            plt.show()
            return StochasticTree(), [], []
    monte_carlo_tree_search(
        stochastic_tree, n_trials=n_trials, max_depth=max_depth, DEBUG=DEBUG, INTERRUPT=INTERRUPT)
    # Planning outputs
    if INTERRUPT is not None:
        if INTERRUPT[0]:
            return None, None, None
    pruned_stochastic_tree = greedy_prune_tree(stochastic_tree)
    optimistic_sequence = optimistic_sequence_from_pruned_tree(pruned_stochastic_tree)
    firstfailure_sequence = firstfailure_then_optimistic_sequence_from_pruned_tree(pruned_stochastic_tree)
    return pruned_stochastic_tree, optimistic_sequence, firstfailure_sequence

def initialize_stochastic_tree(
        state, fixed_state, path_states, path_variants, task_trees, possible_actions, DEBUG=False):
    stochastic_tree = StochasticTree()
    stochastic_tree.fixed_state = fixed_state
    # add initial trees to open nodes
    for path_state, path_variant, task_tree in zip(path_states, path_variants, task_trees):
        startnode = STNode(
            next_id          = stochastic_tree.next_id,  # noqa
            depth            = 0,                        # noqa
            startidx         = 0,                        # noqa
            state            = state,                    # noqa
            path_state       = path_state,               # noqa
            path_variant     = path_variant,             # noqa
            task_tree        = task_tree,                # noqa
            possible_actions = possible_actions,         # noqa
            cumulative_cost  = 0.,                       # noqa
            cumulative_prob  = 1.,                       # noqa
        )
        stochastic_tree.nodes[startnode.id] = startnode
        stochastic_tree.root_nodes.append(startnode.id)
    if DEBUG:
        for _, node in stochastic_tree.nodes:
            print(node.task_tree)
    return stochastic_tree

def expand_stochastic_tree_node(stochastic_tree, nodeid, DEBUG=False):
    fixed_state = stochastic_tree.fixed_state
    node = stochastic_tree.nodes[nodeid]
    if not node.is_open:
        raise ValueError("Requested expansion on a closed node.")
    # Lazy update of the node state, get new task trees and options for the selected node
    is_changed = False
    new_path_state = node.path_state
    new_state = node.state
    if node.task_state_update is not None:
        if DYNAMIC_STATE:
            new_state = node.state.copy()
            is_changed = node.task_state_update.apply_to_state_along_taskpath(new_state)
            if is_changed:
                new_path = node.path_state.path_xy  # stays constant
                new_path_ij = node.path_state.path_ij  # stays constant
                new_path_state = path_state_from_state_and_path(new_state, new_path, new_path_ij)
        elif DYNAMIC_PATH_STATE:
            new_path_state = node.path_state.copy()
            is_changed = node.task_state_update.apply_to_path_state(new_path_state)
    # disable doing the same action again after failure
    less_actions = node.possible_actions
    if not ALLOW_CONSECUTIVE_FAILURE:
        raise NotImplementedError
    # if state has changed, recompute task tree, path_state
    new_path_variant = node.path_variant  # stays constant
    new_task_tree = node.task_tree
    if is_changed:
        task_length_limit = None
        if ACTION_LENGTH_LIMIT_IN_TASK_TREE is not None:
            task_length_limit = int(ACTION_LENGTH_LIMIT_IN_TASK_TREE / fixed_state.map.resolution())
        new_task_tree = task_tree_from_path_state(new_path_state, less_actions,
                                                  task_length_limit, constrain_to_depth_1=True)
    # if goal is reached
    is_terminal = False
    new_startidx = new_path_state.pos_along_path
    if new_startidx == len(new_path_state.path_xy) - 1:
        # create an empty task tree to make the node terminal
        new_task_tree = {new_startidx: TTNode()}
        is_terminal = True
    # update all stale fields
    node.state = new_state
    node.possible_actions = less_actions
    node.path_state = new_path_state
    node.path_variant = new_path_variant
    node.task_tree = new_task_tree
    node.startidx = new_startidx
    node.is_terminal = is_terminal

    # follow first branches, predict outcome, add resulting nodes to tree
    newnodes = []
    first_branches = node.task_tree[node.startidx].childbranches
    for branch in first_branches:
        resultidx, action = branch.target, branch.action
        task_state = CTaskState(node.path_state, node.startidx, resultidx, fixed_state.map.resolution())
        outcomes = action.predict_outcome(task_state, params=None)
        # get resulting state for each outcome
        branch_bundle = BranchBundle(meta={"tasktargetpathidx": resultidx,
                                           "taskaction": action,
                                           "taskfullpath": node.path_state.get_path_xy()})
        for outcome in outcomes:
            # add new node to opennodes and nodes
            # use parent values for stale values which get updated later
            newnode = STNode(
                 next_id          = stochastic_tree.next_id,                          # noqa
                 depth            = node.depth+1,                                     # noqa
                 startidx         = node.startidx,                            # stale # noqa
                 state            = node.state,                               # stale # noqa
                 path_state       = node.path_state,                          # stale # noqa
                 path_variant     = node.path_variant,                        # stale # noqa
                 task_tree        = node.task_tree,                           # stale # noqa
                 possible_actions = node.possible_actions,                    # stale # noqa
                 cumulative_cost  = node.cumulative_cost+outcome.cost,                # noqa
                 cumulative_prob  = node.cumulative_prob*outcome.probability,         # noqa
            )
            newnode.task_state_update = outcome.task_state_update
            stochastic_tree.nodes[newnode.id] = newnode
            newnodes.append(newnode)
            branch_bundle.branches.append(
                Branch(target=newnode.id, action=action,
                       meta={"success": outcome.success,
                             "cost": outcome.cost,
                             "prob": outcome.probability})
            )
        node.childbranchbundles.append(branch_bundle)
    if not node.is_terminal and not node.childbranchbundles:
        warnings.warn("IA: Non terminal node has no branches. Check task trees and preconditions.")
    node.is_open = False
    return newnodes

def build_stochastic_tree(stochastic_tree, DEBUG=False):
    # find open nodes and keep an ranked heap
    def noderank(node):
        # breadth/depth first search
        if node.depth == 0:
            return -np.inf  # always expand root nodes first
#             return np.random.rand()  # random ordering
        return node.depth  # breadth first search
#             return -node.depth  # depth first search
    open_nodes = []
    for nodeid in stochastic_tree.nodes:
        node = stochastic_tree.nodes[nodeid]
        if node.is_open:
            heappush(open_nodes, (noderank(node), node.id))
    # iteratively expand nodes
    itercount = 0
    while open_nodes:
        itercount += 1
        _, nodeid = heappop(open_nodes)
        newnodes = expand_stochastic_tree_node(stochastic_tree, nodeid, DEBUG=DEBUG)
        for newnode in newnodes:
            heappush(open_nodes, (noderank(newnode), newnode.id))
        # stopping criteria
        if itercount > 100:
            break

def monte_carlo_tree_search(stochastic_tree, n_trials=1000, max_depth=10, DEBUG=False, INTERRUPT=None):
    for i in range(n_trials):
        # start a trial
        path_trough_tree = []
        if not stochastic_tree.root_nodes:
            return
        firstnodeidx = stochastic_tree.root_nodes[np.random.randint(
            len(stochastic_tree.root_nodes)
        )]
        node = stochastic_tree.nodes[firstnodeidx]
        # step through the tree, sampling outcomes and expanding nodes until end or max depth
        while True:
            # allow being interrupted externally
            if INTERRUPT is not None:
                if INTERRUPT[0]:
                    return
            # if depth is reached, leave node open, unexpanded
            if node.depth >= max_depth:
                break
            # expand and close open nodes
            if node.is_open:
                expand_stochastic_tree_node(stochastic_tree, node.id, DEBUG=DEBUG)
            # termination conditions
            if node.is_terminal:
                break
            if not node.childbranchbundles:
#                 warnings.warn("Node {} is childless, non-terminal, and closed.".format(node.id))
                break
            # pick action at random
            bundleidx = np.random.randint(len(node.childbranchbundles))
            bundle = node.childbranchbundles[bundleidx]
            path_trough_tree.append((node, bundleidx))  # add action to the path
            # pick outcome stochastically
            choices = [branch.target for branch in bundle.branches]
            probs = np.array([branch.meta["prob"] for branch in bundle.branches], dtype=np.float32)
#                 resultid = np.random.choice(choices, p=probs)  # 10x slower
            resultid = choices[sultans_wife(probs)]
            # update for next loop
            node = stochastic_tree.nodes[resultid]
        # get final cost, add to MC estimate
        endnode = node
        final_is_goal_reached = node.is_terminal
        final_cost = node.cumulative_cost
        if not final_is_goal_reached:
            final_cost = np.nan
        # estimate for first choice, between three paths
        for node, bundleidx in path_trough_tree:
            # create NodeMCEstimateInfo for node if it doesn't have one
            if node.mce_info is None:
                node.mce_info = NodeMCEstimateInfo(
                    [MCBranchEstimate() for _ in node.childbranchbundles])
            estimate = node.mce_info.option_mc_estimates[bundleidx]
            estimate.n_total += 1
            if final_is_goal_reached:
                estimate.n_goal_reached += 1
                estimate.sum_cost_to_goal += final_cost
            if STORE_ESTIMATE_DISTRIBUTIONS:
                estimate.dist.append(final_cost)
        # add final node
        if endnode.mce_info is None:
            endnode.mce_info = NodeMCEstimateInfo(
                [MCBranchEstimate() for _ in endnode.childbranchbundles])
        estimate = endnode.mce_info.endnode_estimate
        estimate.n_total += 1
        if final_is_goal_reached:
            estimate.n_goal_reached += 1
            estimate.sum_cost_to_goal += final_cost
        if STORE_ESTIMATE_DISTRIBUTIONS:
            estimate.dist.append(final_cost)

def greedy_prune_tree(stochastic_tree, DEBUG=False):
    def objective_score(mc_estimate):
        # TODO : pick either lowest cost sequence or mix of lowest cost and low outcome variance
        expcost = mc_estimate.expected_final_cost()
        penalty = mc_estimate.n_total / mc_estimate.n_goal_reached
        return expcost * penalty

    def is_first_score_better(first_score, second_score):
        return first_score <= second_score

    def worst_possible_score():
        return np.inf
    # new pruned trees
    pruned_stochastic_tree = StochasticTree()
    pruned_stochastic_tree.fixed_state = stochastic_tree.fixed_state
    pruned_stochastic_tree.next_id = stochastic_tree.next_id
    # only add best options to pruned trees
    bestnodeid = None
    # find start node
    bestobjectivescore = worst_possible_score()
    for nodeid in stochastic_tree.root_nodes:
        mce_info = stochastic_tree.nodes[nodeid].mce_info
        if mce_info is None:
            continue
        est = mce_info.node_estimate()
        if est.n_goal_reached == 0:
            continue
        objectivescore = objective_score(est)
        if is_first_score_better(objectivescore, bestobjectivescore):
            bestobjectivescore = objectivescore
            bestnodeid = nodeid
    if bestnodeid is None:
        raise ValueError("No estimates found for root nodes")
    # add best root node to pruned trees
    pruned_stochastic_tree.root_nodes = [bestnodeid]
    # add root node to nodes_to_copy
    nodes_to_copy = [bestnodeid]
    # greedy step through tree
    while nodes_to_copy:
        # pick an unprocessed node, copy it into the tree
        nodeid = nodes_to_copy.pop()
        node = stochastic_tree.nodes[nodeid]
        pruned_stochastic_tree.nodes[nodeid] = node.copy()
        # Find best action for this node, add outcomes to nodes_to_copy
        if node.is_terminal or node.is_open:
            continue
        mce_info = node.mce_info
        if mce_info is None:
            print("ERROR: closed non terminal node with no mce_info")
            continue
        estimates = mce_info.option_mc_estimates
        bestobjectivescore = worst_possible_score()
        bestoptionidx = None
        for i, est in enumerate(estimates):
            if est.n_goal_reached == 0:
                continue
            objectivescore = objective_score(est)
            if is_first_score_better(objectivescore, bestobjectivescore):
                bestobjectivescore = objectivescore
                bestoptionidx = i
        if bestoptionidx is None and estimates:
            # there are branches but none of them lead to the goal. pick any explored one.
            for i, est in enumerate(estimates):
                if est.n_total > 0:
                    bestoptionidx = i
        if bestoptionidx is None:
            continue
        bestoptionbundle = node.childbranchbundles[bestoptionidx]
        # leave only best option in the stochastic tree
        pruned_stochastic_tree.nodes[nodeid].childbranchbundles = [bestoptionbundle]
        pruned_stochastic_tree.nodes[nodeid].mce_info.option_mc_estimates = [
            node.mce_info.option_mc_estimates[bestoptionidx]
        ]
        # add all option outcomes to nodes_to_copy
        bundle = bestoptionbundle
        for branch in bundle.branches:
            nodes_to_copy.append(branch.target)
    # result
    return pruned_stochastic_tree

def select_best_sequence_from_mcts_estimates(stochastic_tree, DEBUG=False):
    def objective_score(mc_estimate):
        # TODO : pick either lowest cost sequence or mix of lowest cost and low outcome variance
        return mc_estimate.expected_final_cost()

    def is_first_score_better(first_score, second_score):
        return first_score <= second_score

    def worst_possible_score():
        return np.inf
    sequence = []
    bestnodeid = None
    # find start node
    bestobjectivescore = worst_possible_score()
    for nodeid in stochastic_tree.root_nodes:
        mce_info = stochastic_tree.nodes[nodeid].mce_info
        if mce_info is None:
            continue  # we don't have monte carlo info on this node
        est = mce_info.node_estimate()
        if est.n_goal_reached == 0:
            continue
        objectivescore = objective_score(est)
        if is_first_score_better(objectivescore, bestobjectivescore):
            bestobjectivescore = objectivescore
            bestnodeid = nodeid
        if DEBUG:
            print("ROOT NODE {} ---------------".format(nodeid))
            print(est.n_goal_reached)
            print(est.n_total)
            print(est.sum_cost_to_goal)
            p = None
            objectivescore = None
            if est.n_total > 0:
                p = est.prob_reach_goal()
            if est.n_goal_reached > 0:
                objectivescore = objective_score(est)
            print("p: ", p)
            print("objectivescore: ", objectivescore)
    if bestnodeid is None:
        raise ValueError("No estimates found for root nodes")
    # greedy step through tree
    nodeid = bestnodeid
    while True:
        if DEBUG:
            print("RESULTING NODE: {} ---------------".format(nodeid))
        if stochastic_tree.nodes[nodeid].is_terminal:
            if DEBUG:
                print("GOAL REACHED")
            break
        estimates = stochastic_tree.nodes[nodeid].mce_info.option_mc_estimates
        bestobjectivescore = worst_possible_score()
        bestoptionidx = None
        for i, est in enumerate(estimates):
            if est.n_goal_reached == 0:
                continue
            objectivescore = objective_score(est)
            if is_first_score_better(objectivescore, bestobjectivescore):
                bestobjectivescore = objectivescore
                bestoptionidx = i
            if DEBUG:
                print("-- Option -- ")
                print(est.n_goal_reached)
                print(est.n_total)
                print(est.sum_cost_to_goal)
                p = None
                objectivescore = None
                if est.n_total > 0:
                    p = est.prob_reach_goal()
                if est.n_goal_reached > 0:
                    objectivescore = objective_score(est)
                print("p: ", p)
                print("objectivescore: ", objectivescore)
        if bestoptionidx is None:
            break
        if DEBUG:
            print("SELECTED OPTION -------> {}".format(bestoptionidx))
        sequence.append((nodeid, bestoptionidx))
        # follow option stochastically
        bundle = stochastic_tree.nodes[nodeid].childbranchbundles[bestoptionidx]
        choices = [branch.target for branch in bundle.branches]
        probs = [branch.meta["prob"] for branch in bundle.branches]
        nodeid = np.random.choice(choices, p=probs)
    # result
    return sequence

def optimistic_sequence_from_pruned_tree(pruned_stochastic_tree):
    sequence = []
    # start node
    nodeid = pruned_stochastic_tree.root_nodes[0]
    # step through tree, picking optimistic outcomes
    while True:
        if nodeid not in pruned_stochastic_tree.nodes:
            break
        if pruned_stochastic_tree.nodes[nodeid].is_terminal:
            break
        bestoptionidx = 0
        # follow option stochastically
        if not pruned_stochastic_tree.nodes[nodeid].childbranchbundles:
            break
        bundle = pruned_stochastic_tree.nodes[nodeid].childbranchbundles[bestoptionidx]
        successbranches = [branch for branch in bundle.branches if branch.meta["success"] == 1]
        successprobs = np.array([branch.meta["prob"] for branch in successbranches])
        successprobs = successprobs / np.sum(successprobs)  # normalize to sum = 1
        branch = np.random.choice(successbranches, p=successprobs)
        nodeid = branch.target
        sequence.append((bundle.meta, branch.meta))
    # result
    return sequence

def firstfailure_then_optimistic_sequence_from_pruned_tree(pruned_stochastic_tree):
    sequence = []
    # start node
    nodeid = pruned_stochastic_tree.root_nodes[0]
    # step through tree, picking optimistic outcomes
    first = True
    while True:
        if nodeid not in pruned_stochastic_tree.nodes:
            break
        if pruned_stochastic_tree.nodes[nodeid].is_terminal:
            break
        if not pruned_stochastic_tree.nodes[nodeid].childbranchbundles:
            break
        bestoptionidx = 0
        # follow option stochastically
        bundle = pruned_stochastic_tree.nodes[nodeid].childbranchbundles[bestoptionidx]
        if first:
            failurebranches = [branch for branch in bundle.branches if branch.meta["success"] < 1]
            failureprobs = np.array([branch.meta["prob"] for branch in failurebranches])
            failureprobs = failureprobs / np.sum(failureprobs)  # normalize to sum = 1
            branch = np.random.choice(failurebranches, p=failureprobs)
            first = False
        else:
            successbranches = [branch for branch in bundle.branches if branch.meta["success"] == 1]
            successprobs = np.array([branch.meta["prob"] for branch in successbranches])
            successprobs = successprobs / np.sum(successprobs)  # normalize to sum = 1
            branch = np.random.choice(successbranches, p=successprobs)
        nodeid = branch.target
        sequence.append((bundle.meta, branch.meta))
    # result
    return sequence

# UTILITY FUNCTIONS --------
def permutations_from_task_tree(task_tree):
    permutations = []

    def expandnode(node, task_tree, running_tasklist, permutations):
        # check if end reached
        child_branches = task_tree[node].childbranches
        if not child_branches:
            permutations.append(running_tasklist)
        for branch in child_branches:
            # prevent doing the same action over and over at same node (inf recursion)
            if node == branch.target:
                if branch.action in [task.action for task in running_tasklist]:
                    continue
            expandnode(branch.target, task_tree, running_tasklist + [branch, ], permutations)
    expandnode(0, task_tree, [], permutations)
    return permutations

# VISUALIZATION & OUTPUT ----
def visualize_task(action, path_ij):
    pass

def visualize_task_tree(task_tree):
    # assign value 1-N to each node in tree
    nodevals = {}
    for i, node in enumerate(task_tree):  # can also use sorted(task_tree)
        nodevals[node] = i
    # set of actions in tree
    actionset = set()
    for node in task_tree:
        branches = task_tree[node].childbranches
        for branch in branches:
            action = branch.action
            actionset.add(action)
    # assign float value 1-N to each action
    actionvals = {}
    for i, action in enumerate(actionset):
        actionvals[action] = i
    # plot tree
    legendcount = [0 for _ in actionset]  # hack to tie each action legend to a single branch
    for node in task_tree:
        node_x = nodevals[node]
        node_y = node
        plt.scatter([node_x], [node_y], color=(1, 1, 1, 1), edgecolors=(0, 0, 0, 1))
        branches = task_tree[node].childbranches
        for branch in branches:
            actionval = actionvals[branch.action]
            label = None
            if not legendcount[actionval]:
                label = branch.action.typestr()
                legendcount[actionval] = 1
            result_x = nodevals[branch.target]
            result_y = branch.target
            plt.plot([node_x, result_x], [node_y, result_y], ',-',
                     color=branch.action.color(), label=label)
    plt.gca().invert_yaxis()
    plt.gca().set_xticklabels([node for i, node in enumerate(task_tree)])
    plt.xticks([i for i, node in enumerate(task_tree)])
    plt.gca().set_yticklabels([node for i, node in enumerate(task_tree)])
    plt.yticks([node for i, node in enumerate(task_tree)])
    plt.legend()

def visualize_stochastic_tree(stochastic_tree, x_order="id"):
    # coordinate definitions
    if x_order == "id":
        def _x(node):
            return node.id
    elif x_order == "progress":
        def _x(node):
            return node.startidx
    elif x_order == "cost":
        def _x(node):
            return node.cumulative_cost
    elif x_order == "probability":
        def _x(node):
            return node.cumulative_prob
    else:
        raise NotImplementedError

    def _y(node):
        return node.depth
    # set of actions in tree
    actionset = set()
    for nodeid in stochastic_tree.nodes:
        bundles = stochastic_tree.nodes[nodeid].childbranchbundles
        for bundle in bundles:
            for branch in bundle.branches:
                actionset.add(branch.action)
    # assign float value 1-N to each action
    actionvals = {}
    for i, action in enumerate(actionset):
        actionvals[action] = i
    # plot tree
    legendcount = [0 for _ in actionset]  # hack to tie each action legend to a single branch
    for nodeid in stochastic_tree.nodes:
        node = stochastic_tree.nodes[nodeid]
        node_x = _x(node)
        node_y = _y(node)
        # draw nodes with no parents
        if node.depth == 0:
            plt.scatter([node_x], [node_y], color=(1, 1, 1, 1), edgecolors=(0, 0, 0, 1))
        # don't draw open node branches as they lead nowhere
        if node.is_open:
            continue
        # draw node branches
        bundles = node.childbranchbundles
        for bundle in bundles:
            branches = bundle.branches
            # find bundle midpoint
            mid_x = 0
            mid_y = 0
            for branch in branches:
                target_node = stochastic_tree.nodes[branch.target]
                result_x = _x(target_node)
                result_y = _y(target_node)
                mid_x += result_x
                mid_y += result_y
            mid_x = 1. * mid_x / len(branches)
            mid_y = 1. * mid_y / len(branches)
            fork_x = node_x + 0.5 * (mid_x - node_x)
            fork_y = node_y + 0.5 * (mid_y - node_y)
            # draw branches in bundle
            for branch in branches:
                target_node = stochastic_tree.nodes[branch.target]
                result_x = _x(target_node)
                result_y = _y(target_node)
                actionval = actionvals[branch.action]
                label = None
                if not legendcount[actionval]:
                    label = branch.action.typestr()
                    legendcount[actionval] = 1
                successcolor = "green" if branch.meta["success"] else "red"
                actioncolor = branch.action.color()
                # Y shaped branch
                plt.plot([node_x, fork_x, result_x], [node_y, fork_y, result_y], ',-',
                         color=actioncolor, label=label)
                if target_node.is_terminal:
                    plt.scatter([result_x], [result_y],
                                marker="*", color=(1, 1, 0, 1), edgecolors=(0.4, 0.4, 0, 1))
                else:
                    _ = "o" if target_node.is_open else "s"
                    plt.scatter([result_x], [result_y], color=(1, 1, 1, 1), edgecolors=successcolor)

    plt.gca().invert_yaxis()
    xticks = [nodeid for nodeid in stochastic_tree.nodes]
    if False:
        plt.gca().set_xticklabels(xticks)
        plt.xticks(xticks)
    # set of y ticks
    yticks = [stochastic_tree.nodes[nodeid].depth for nodeid in stochastic_tree.nodes]
    plt.gca().set_yticklabels(yticks)
    plt.yticks(yticks)
    plt.legend()
    plt.xlabel(x_order)
    plt.ylabel("depth")

def visualize_mc_estimate_tree(stochastic_tree, x_order="expcost", hide_unexplored=True):
    # coordinate definitions
    if x_order == "n_total":
        def _x(node):
            return node.mce_info.node_estimate().n_total
    elif x_order == "expcost":
        def _x(node):
            if node.mce_info is None:  # open node, not reached by MC
                # treat mce same as terminal node for visualization purposes
                return node.cumulative_cost
            x = node.mce_info.node_estimate().expected_final_cost()
            if np.isnan(x):
                x = node.cumulative_cost
            return x

    def _y(node):
        return node.depth
    # total trials
    max_width = 10
    n_total_trials = 0
    for nodeid in stochastic_tree.root_nodes:
        mce_info = stochastic_tree.nodes[nodeid].mce_info
        if mce_info is None:
            continue
        n_total_trials += mce_info.node_estimate().n_total
    # set of actions in tree
    actionset = set()
    for nodeid in stochastic_tree.nodes:
        bundles = stochastic_tree.nodes[nodeid].childbranchbundles
        for bundle in bundles:
            for branch in bundle.branches:
                actionset.add(branch.action)
    # set of path variants in tree
    pathvariantsset = set()
    for nodeid in stochastic_tree.root_nodes:
        pathvariantsset.add(stochastic_tree.nodes[nodeid].path_variant)
    # assign float value 1-N to each action
    actionvals = {}
    for i, action in enumerate(actionset):
        actionvals[action] = i
    # hack to tie each action legend to a single branch
    legendcount = [0 for _ in actionset]
    # assign success to each node
    for nodeid in stochastic_tree.nodes:
        stochastic_tree.nodes[nodeid].meta_success = None
        stochastic_tree.nodes[nodeid].meta_parent = None
    for nodeid in stochastic_tree.nodes:
        node = stochastic_tree.nodes[nodeid]
        bundles = node.childbranchbundles
        for i, bundle in enumerate(bundles):
            branches = bundle.branches
            for branch in branches:
                if branch.target not in stochastic_tree.nodes:
                    continue
                stochastic_tree.nodes[branch.target].meta_success = branch.meta["success"]
                stochastic_tree.nodes[branch.target].meta_parent = node.id
    # plot tree
    scatterpoints = {'x': [], 'y': [], 'c': [], 'edgecolors': []}
    terminalpoints = {'x': [], 'y': [], 'c': [], 'edgecolors': []}
    scatternodeidmap = {}
    # draw nodes
    for nodeid in stochastic_tree.nodes:
        node = stochastic_tree.nodes[nodeid]
        node_x = _x(node)
        node_y = _y(node)
        # node
        nodecolor = (1., 1., 1., 1.)  # white
        edgecolor = "green" if node.meta_success else "red"
        if node.depth == 0 or node.meta_success is None:
            edgecolor = "black"
        if node.mce_info is None:
            edgecolor = "grey"
            if hide_unexplored:
                continue
        if node.is_terminal:
            terminalpoints['x'].append(node_x)
            terminalpoints['y'].append(node_y)
            terminalpoints['c'].append((1., 1., 0., 1.))  # yellow
            terminalpoints['edgecolors'].append("black")
            nodecolor = (1., 1., 1., 1.)
            edgecolor = "white"
        scatterpoints['x'].append(node_x)
        scatterpoints['y'].append(node_y)
        scatterpoints['c'].append(nodecolor)
        scatterpoints['edgecolors'].append(edgecolor)
        scatternodeidmap[(node_x, node_y)] = node.id
    # draw node branches
    for nodeid in stochastic_tree.nodes:
        node = stochastic_tree.nodes[nodeid]
        node_x = _x(node)
        node_y = _y(node)
        bundles = node.childbranchbundles
        for i, bundle in enumerate(bundles):
            branches = bundle.branches
            # find bundle midpoint
            mid_x = 0
            mid_y = 0
            n_tines = 0  # fork tines
            for branch in branches:
                if branch.target not in stochastic_tree.nodes:
                    continue
                target_node = stochastic_tree.nodes[branch.target]
                if target_node.mce_info is None:
                    if hide_unexplored:
                        continue
                result_x = _x(target_node)
                result_y = _y(target_node)
                mid_x += result_x
                mid_y += result_y
                n_tines += 1
            fork_x = node_x
            fork_y = node_y
            if n_tines > 0:
                mid_x = 1. * mid_x / n_tines
                mid_y = 1. * mid_y / n_tines
                fork_x = node_x + 0.5 * (mid_x - node_x)
                fork_y = node_y + 0.5 * (mid_y - node_y)
            # draw branches in bundle
            for branch in branches:
                if branch.target not in stochastic_tree.nodes:
                    print("WARNING: stochastic tree has ghost branches (from {}, to {})".format(
                        node.id, branch.target))
                    plt.scatter(node_x, node_y, facecolors="black", marker="o", zorder=3)
                    print(node_x, node_y)
                    continue
                target_node = stochastic_tree.nodes[branch.target]
                result_x = _x(target_node)
                result_y = _y(target_node)
                actionval = actionvals[branch.action]
                label = None
                if not legendcount[actionval]:
                    label = branch.action.typestr()
                    legendcount[actionval] = 1
                actioncolor = branch.action.color()
                # Y shaped branch # linewidth varies with trials in bundle estimate
                # sqrt is used to make most comparable
                if target_node.mce_info is None:
                    if hide_unexplored:
                        continue
                    else:
                        linewidth = 0
                else:
                    linewidth = np.sqrt(
                        1. * target_node.mce_info.node_estimate().n_total / n_total_trials) * max_width
                linewidth = max(1, linewidth)
                linestyle = target_node.path_variant.linestyle()
                plt.plot([node_x, fork_x, result_x], [node_y, fork_y, result_y],
                         linestyle=linestyle, color=actioncolor,
                         label=label, linewidth=linewidth, zorder=1)
    # plot nodes
    sc = plt.scatter(zorder=2, marker="o", **scatterpoints)
    plt.scatter(zorder=2, marker="*", **terminalpoints)
    # Formatting
    plt.gca().invert_yaxis()
    xticks = [nodeid for nodeid in stochastic_tree.nodes]
    if False:
        plt.gca().set_xticklabels(xticks)
        plt.xticks(xticks)
    # set of y ticks
    yticks = [stochastic_tree.nodes[nodeid].depth for nodeid in stochastic_tree.nodes]
    plt.gca().set_yticklabels(yticks)
    plt.yticks(yticks)
    plt.legend()
    plt.xlabel(x_order)
    plt.ylabel("depth")
    plt.title("Monte Carlo estimates tree. {}".format(", ".join(["'{}' {}".format(
        variant.linestyle(), variant.typestr()) for variant in pathvariantsset])))
    # All this to have a live tooltip
    ax = plt.gca()
    fig = plt.gcf()
    annot = ax.annotate("", xy=(0, 0), xytext=(20, 20), textcoords="offset points",
                        bbox=dict(boxstyle="round", fc="w"),
                        arrowprops=dict(arrowstyle="->"))
    annot.set_visible(False)

    def update_annot(ind):
        scindex = ind["ind"][0]
        pos = sc.get_offsets()[scindex]
        annot.xy = pos
        nodeindex = scatternodeidmap[(pos[0], pos[1])]
        node = stochastic_tree.nodes[nodeindex]
        children = [[branch.target for branch in bundle.branches] for bundle in node.childbranchbundles]
        notintree = [child for outcomes in children for child in outcomes
                     if child not in stochastic_tree.nodes]
        est = node.mce_info.node_estimate().expected_final_cost() if node.mce_info is not None else -1
        text = "node {}, depth {}, p {:.2f}, est {:.2f}, parent {}, children {}, n.i.t {}".format(
            node.id, node.depth, node.cumulative_prob, est, node.meta_parent, children, notintree)
        annot.set_text(text)
        annot.get_bbox_patch().set_facecolor(scatterpoints['edgecolors'][scindex])
        annot.get_bbox_patch().set_alpha(0.4)

    def hover(event):
        vis = annot.get_visible()
        if event.inaxes == ax:
            cont, ind = sc.contains(event)
            if cont:
                update_annot(ind)
                annot.set_visible(True)
                fig.canvas.draw_idle()
            else:
                if vis:
                    annot.set_visible(False)
                    fig.canvas.draw_idle()
    fig.canvas.mpl_connect("motion_notify_event", hover)
    # ---

def visualize_mc_estimate_distribution(estimate, n_bins=100, color=None, goal_not_reached_val=-1):
    if not STORE_ESTIMATE_DISTRIBUTIONS:
        raise NotImplementedError
    distribution = np.array(estimate.dist)
    distavg = np.mean(distribution[np.logical_not(np.isnan(distribution))])
    distribution[np.isnan(distribution)] = goal_not_reached_val
    plt.hist(distribution, bins=n_bins, color=color,
             histtype='stepfilled', alpha=0.3, density=True, ec="k")
    plt.axvline(distavg, color=color)
    plt.xlabel("expected cost")
    plt.ylabel("proportion of samples in bin")
    plt.title("Distribution of Expected Cost for Node")

def visualize_root_node_mc_estimate_distributions(stochastic_tree, n_bins=100, goal_not_reached_val=-1):
    legend = []
    for nodeid in stochastic_tree.root_nodes:
        node = stochastic_tree.nodes[nodeid]
        mce_info = node.mce_info
        if mce_info is None:
            continue
        visualize_mc_estimate_distribution(mce_info.node_estimate(),
                                           n_bins=n_bins,
                                           goal_not_reached_val=goal_not_reached_val,
                                           color=node.path_variant.linecolor())
        legend.append(node.path_variant.typestr())
    plt.legend(legend)
    plt.title("Distribution of Expected Cost for Root Nodes")

def visualize_task_permutations(task_permutations):
    # set of nodes
    nodeset = {0}
    for p in task_permutations:
        for task in p:
            nodeset.add(task.target)
    # set of actions in perm
    actionset = set()
    for p in task_permutations:
        for task in p:
            action = task.action
            actionset.add(action)
    # assign float value 1-N to each action
    actionvals = {}
    for i, action in enumerate(actionset):
        actionvals[action] = i
    legendcount = [0 for _ in actionset]
    for i, p in enumerate(task_permutations):
        prev = 0
        for task in p:
            actionval = actionvals[task.action]
            label = None
            if not legendcount[actionval]:
                label = task.action.typestr()
                legendcount[actionval] = 1
            color = task.action.color()
            plt.bar(i, [task.target - prev], bottom=[prev],
                    color=color, label=label,
                    edgecolor=(0, 0, 0, 1))
            prev = task.target
    plt.gca().invert_yaxis()
    plt.gca().set_xticklabels([])
    plt.xticks([])
    plt.gca().set_yticklabels([node for node in nodeset])
    plt.yticks([node for node in nodeset])
    plt.legend()

def visualize_action_preconditions(path_state, possible_actions):
    for i, action in enumerate(possible_actions):
        proto_segments, _ = action.check_preconditions_on_path(path_state)
        color = action.color()
        bool1d = bool1d_from_segments(proto_segments, len(path_state.get_path_xy()))
        plt.plot(bool1d, 's--', color=color)
    plt.legend([action.typestr() for action in possible_actions])

def visualize_dijkstra_field(state, fixed_state, path_variant):
    dijkstra = fixed_state.derived_state.dijkstra_from_goal(state, fixed_state, path_variant)
    gridshow(dijkstra)
    plt.title(path_variant.typestr() + "_dijkstra")

def visualize_dijkstra_fields(state, fixed_state, path_variants):
    n = len(path_variants)
    nrow = n
    ncol = 1
    fig, axes = plt.subplots(nrow, ncol)
    if not isinstance(axes, np.ndarray):
        axes = np.array([axes])
    for axrow, path_variant in zip(axes, path_variants):
        plt.sca(axrow)
        visualize_dijkstra_field(state, fixed_state, path_variant)

def visualize_traversability(state, fixed_state, path_variant):
    is_traversable = path_variant.is_state_traversable(state, fixed_state)
    gridshow(is_traversable.astype(int))
    plt.title(path_variant.typestr() + "_is_traversable")

def visualize_traversabilities(state, fixed_state, path_variants):
    n = len(path_variants)
    nrow = n
    ncol = 1
    fig, axes = plt.subplots(nrow, ncol)
    if not isinstance(axes, np.ndarray):
        axes = np.array([axes])
    for axrow, path_variant in zip(axes, path_variants):
        plt.sca(axrow)
        visualize_traversability(state, fixed_state, path_variant)

def visualize_state_feature(state, fixed_state, feature_name, hide_uncertain=True, uncertainties=False):
    if uncertainties:
        grid = 1. * state.grid_features_uncertainties()[kStateFeatures[feature_name], :]
    else:
        grid = 1. * state.grid_features_values()[kStateFeatures[feature_name], :]
    grid = np.copy(grid)
    if hide_uncertain and not uncertainties:
        grid[np.greater_equal(state.grid_features_uncertainties()[kStateFeatures[feature_name], :],
                              1.)] = np.nan
    grid[fixed_state.map.occupancy() > fixed_state.map.thresh_occupied()] = np.nan
    # color scale
    vmin = 0
    vmax = 1.
    if feature_name == "crowdedness":
        vmax = 6.
    if uncertainties:
        vmax = 2.
    gs = gridshow(grid, cmap=plt.cm.Greens, vmin=vmin, vmax=vmax)
    ij = state.get_pos_ij()
    plt.scatter(ij[0], ij[1], marker='+', c="black")
    uncertainty_str = " uncert." if uncertainties else ""
    plt.title(feature_name + uncertainty_str)
    return gs

def visualize_state_features(state, fixed_state, hide_uncertain=False):
    n = len(list(kStateFeatures))
    nrow = n
    ncol = 2
    fig, axes = plt.subplots(nrow, ncol)
    for axrow, feature_name in zip(axes, kStateFeatures):
        plt.sca(axrow[0])
        visualize_state_feature(
            state, fixed_state, feature_name, hide_uncertain=hide_uncertain, uncertainties=False)
        plt.sca(axrow[1])
        visualize_state_feature(
            state, fixed_state, feature_name, hide_uncertain=hide_uncertain, uncertainties=True)

def visualize_world(state_e, fixed_state, possible_actions, env,
                    fig=None, possible_path_variants=None):
    paths_xy, path_variants = path_options(state_e, fixed_state,
                                           possible_path_variants=possible_path_variants)
    # ---------- PLOTTING
    if fig is None:
        fig = plt.figure("world")
        plt.cla()
    contours = fixed_state.map.as_closed_obst_vertices()
    fixed_state.map.plot_contours(contours, '-k')
    env.plot_agents_xy()
    for path_xy, variant in zip(paths_xy, path_variants):
        plt.plot(path_xy[:, 0], path_xy[:, 1], color=variant.linecolor())
    plt.axis('equal')

def visualize_simple(state_e, fixed_state, pruned_stochastic_tree, byproducts):
    # From byproducts
    path_variants = byproducts["path_variants"]
    path_states = byproducts["path_states"]
    # plot
    visualize_state_feature(state_e, fixed_state, "crowdedness", hide_uncertain=True)
    # paths
    for i, (path_state, path_variant) in enumerate(zip(
            path_states, path_variants)):
        path_ij = path_state.get_path_ij()
        plt.plot(path_ij[:, 0], path_ij[:, 1], '--', color=path_variant.linecolor())
    # chosen path
    if pruned_stochastic_tree.root_nodes:
        node = pruned_stochastic_tree.nodes[pruned_stochastic_tree.root_nodes[0]]
        chosen_path_variant = node.path_variant
        chosen_path_state = node.path_state
        path_ij = chosen_path_state.get_path_ij()
        plt.plot(path_ij[:, 0], path_ij[:, 1], color=chosen_path_variant.linecolor())

def visualize_sequence_as_path(optimistic_sequence, fixed_state):
    prev_idx = 0
    for task, cps in optimistic_sequence:
        target_idx = task["tasktargetpathidx"]
        action = task["taskaction"]
        path_ij = fixed_state.map.xy_to_floatij(task["taskfullpath"][prev_idx:target_idx])
        plt.plot(path_ij[:, 0], path_ij[:, 1], color=action.color())
        plt.scatter(path_ij[-1, 0], path_ij[-1, 1], color='k', marker='.', zorder=10)
        prev_idx = target_idx

def visualize_planning_process(state_e, fixed_state, possible_actions,
                               pruned_stochastic_tree, optimistic_sequence, firstfailure_sequence,
                               byproducts, env=None):
    # ----------------- PROCESSING -------------------------------------------------------------------
    # From byproducts
    path_variants = byproducts["path_variants"]
    path_states = byproducts["path_states"]
    task_trees = byproducts["task_trees"]
    stochastic_tree = byproducts["stochastic_tree"]
    # ---------------- PLOTTING ----------------------------------------------------------------------
    visualize_state_features(state_e, fixed_state, hide_uncertain=False)
    visualize_traversabilities(state_e, fixed_state, paths.PATH_VARIANTS)
    visualize_dijkstra_fields(state_e, fixed_state, paths.PATH_VARIANTS)
    if env is not None:
        visualize_world(state_e, fixed_state, possible_actions, env)
    # Path Options & Task Trees
    for i, (path_state, path_variant, tree) in enumerate(zip(
            path_states, path_variants, task_trees)):
        path_xy = path_state.get_path_xy()
        path_ij = path_state.get_path_ij()
        fig, axes = plt.subplots(3, 2, num=path_variant.typestr())
        plt.sca(axes[0, 0])
        plt.title("permissivity_map")
        visualize_state_feature(state_e, fixed_state, "permissivity")
        plt.plot(path_ij[:, 0], path_ij[:, 1], color=path_variant.linecolor())
        plt.sca(axes[0, 1])
        plt.title("path in world")
        contours = fixed_state.map.as_closed_obst_vertices()
        fixed_state.map.plot_contours(contours, '-k')
        if env is not None:
            env.plot_agents_xy()
        plt.plot(path_xy[:, 0], path_xy[:, 1], color=path_variant.linecolor())
        plt.axis('equal')
        plt.sca(axes[1, 0])
        plt.title("state features along path")
        for feature_name in kStateFeatures:
            n = kStateFeatures[feature_name]
            values = path_state.path_features_values()[n, :]
            uncertainties = path_state.path_features_uncertainties()[n, :]
            max_y = max(np.max(uncertainties), np.max(values))
            plt.plot(values, 's--', label=feature_name, alpha=0.5)
            plt.plot(uncertainties, 'o--', label=feature_name + "_uncertainty", alpha=0.5)
            plt.ylim([0, min(max_y, 5.)])
        plt.xlim([0, len(path_state.get_path_xy())])
        plt.legend()
        plt.sca(axes[1, 1])
        plt.title("task tree for path")
        visualize_task_tree(tree)
        plt.sca(axes[2, 0])
        plt.title("action preconditions along path")
        visualize_action_preconditions(path_state, possible_actions)  # SEGFAULT?
        plt.sca(axes[2, 1])
        plt.title("task permutations for path")
        permutations = permutations_from_task_tree(tree)
        visualize_task_permutations(permutations)
    plt.figure("mc_estimate_tree")
    visualize_mc_estimate_tree(stochastic_tree)
    plt.figure("pruned_tree")
    visualize_mc_estimate_tree(pruned_stochastic_tree)
    plt.figure("outcome_distributions")
    visualize_root_node_mc_estimate_distributions(stochastic_tree)
    plot_closeall_button()

def print_tree(task_tree):
    for node in task_tree:
        print("---------- {} ---------".format(node))
        branches = task_tree[node].childbranches
        print([b[0] for b in branches])

def print_mcts_estimates(stochastic_tree):
    for nodeid in stochastic_tree.root_nodes:
        mce_info = stochastic_tree.nodes[nodeid].mce_info
        print("ROOT NODE {} ---------------".format(nodeid))
        if mce_info is None:
            print("no MC info")
            continue
        est = mce_info.node_estimate()
        print(est.n_goal_reached)
        print(est.n_total)
        print(est.sum_cost_to_goal)
        p = None
        expcost = None
        if est.n_total > 0:
            p = est.prob_reach_goal()
        if est.n_goal_reached > 0:
            expcost = est.expected_final_cost()
        print("p: ", p)
        print("expcost: ", expcost)
    for nodeid in stochastic_tree.nodes:
        mce_info = stochastic_tree.nodes[nodeid].mce_info
        print("NODE {} ---------------".format(nodeid))
        if mce_info is None:
            print("no MC info")
            continue
        estimates = mce_info.option_mc_estimates
        for est in estimates:
            print("-- Action -- ")
            print(est.n_goal_reached)
            print(est.n_total)
            print(est.sum_cost_to_goal)
            p = None
            expcost = None
            if est.n_total > 0:
                p = est.prob_reach_goal()
            if est.n_goal_reached > 0:
                expcost = est.expected_final_cost()
            print("p: ", p)
            print("expcost: ", expcost)



class Node(object):
    def __init__(self, worldstate, neighbors, edges, rank, values):
        self.worldstate = worldstate
        self.neighbors = neighbors
        self.edges = edges
        self.rank = rank
        self.values = values



# distutils: language=c++

from libcpp cimport bool
from libcpp.queue cimport priority_queue as cpp_priority_queue
from libcpp.pair cimport pair as cpp_pair
import numpy as np
import copy
cimport numpy as np
from cython.operator cimport dereference as deref
cimport cython
from math import sqrt
from libc.math cimport cos as ccos
from libc.math cimport sin as csin
from libc.math cimport acos as cacos
from libc.math cimport sqrt as csqrt
from libc.math cimport floor as cfloor

import matplotlib.pyplot as plt
from CMap2D import path_from_dijkstra_field
from CMap2D import CMap2D
from map2d import gridshow

# compile time feature definition
DEF N_STATE_FEATURES = 3
DEF CROWDEDNESS_INDEX = 0
DEF PERMISSIVITY_INDEX = 1
DEF PERCEPTIVITY_INDEX = 2
DEF CROWDEDNESS_NAME = u"crowdedness"
DEF PERMISSIVITY_NAME = u"permissivity"
DEF PERCEPTIVITY_NAME = u"perceptivity"

DEF PROB_DISTANCE_REF = 10.

# feature definition (exported to outside libraries)
kStateFeatures = {
    CROWDEDNESS_NAME: CROWDEDNESS_INDEX,
    PERMISSIVITY_NAME: PERMISSIVITY_INDEX,
    PERCEPTIVITY_NAME: PERCEPTIVITY_INDEX,
} # flow, agency, ...


# Features are represented as normal distributions at every point in space
# Ensemble of state features over N-dimensional space
cdef class CVectorFeatures:
    cdef np.float32_t[:,::1] values
    cdef np.float32_t[:,::1] uncertainties
    def __cinit__(self, int task_path_length):
        self.values = np.zeros((N_STATE_FEATURES, task_path_length), dtype=np.float32)
        self.uncertainties = np.inf * np.ones((N_STATE_FEATURES, task_path_length), dtype=np.float32)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    @cython.cdivision(True)
    cdef void copy_from(self, CVectorFeatures source):
        self.values[:,:] = source.values
        self.uncertainties[:,:] = source.uncertainties
    cdef np.float32_t[:,::1] _values(self):
        return self.values
    cdef np.float32_t[:,::1] _uncertainties(self):
        return self.uncertainties
cdef class CGridFeatures:
    cdef np.float32_t[:,:,::1] values
    cdef np.float32_t[:,:,::1] uncertainties
    cdef int grid_width
    cdef int grid_height
    def __cinit__(self, int state_map_width, int state_map_height):
        self.grid_width = state_map_width
        self.grid_height = state_map_height
        self.values = np.zeros(
            (N_STATE_FEATURES, state_map_width, state_map_height), dtype=np.float32)
        self.uncertainties = np.inf * np.ones(
            (N_STATE_FEATURES, state_map_width, state_map_height), dtype=np.float32)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    @cython.cdivision(True)
    cdef void copy_from(self, CGridFeatures source):
        self.values[:,:,:] = source.values
        self.uncertainties[:,:,:] = source.uncertainties
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    @cython.cdivision(True)
    cdef void fill(self, np.float32_t[:,:,::1] values, np.float32_t[:,:,::1] uncertainties):
        cdef int i
        cdef int j
        cdef int n
        cdef int max_n = N_STATE_FEATURES
        for i in range(self.grid_width):
            for j in range(self.grid_height):
                for n in range(max_n):
                    self.values[n, i, j] = values[n, i, j]
                    self.uncertainties[n, i, j] = uncertainties[n, i, j]
    cdef np.float32_t[:,:,::1] _values(self):
        return self.values
    cdef np.float32_t[:,:,::1] _uncertainties(self):
        return self.uncertainties
    def serialize(self):
        return {
            "values": np.array(self.values),
            "uncertainties": np.array(self.uncertainties),
            "grid_width": int(self.grid_width),
            "grid_height": int(self.grid_height),
        }
def unserialize_cgridfeatures(dict_):
    gf = CGridFeatures(dict_["grid_width"], dict_["grid_height"])
    gf.fill(dict_["values"], dict_["uncertainties"])
    return gf

# Feature diffs stores the operation to be executed on a feature
ctypedef np.float32_t (*Func)(np.float32_t[:] values, np.float32_t[:] uncertainties)
cdef class CFeaturesDiff:
    cdef public bool is_changed[N_STATE_FEATURES]
    cdef Func value_operations[N_STATE_FEATURES]
    cdef Func uncertainty_operations[N_STATE_FEATURES]
    def __cinit__(self):
        cdef int i
        for i in range(N_STATE_FEATURES):
            self.is_changed[i] = False

# The following classes represent the state at various levels of locality
cdef class CPathState:
    """ Reduction of State: 2d elements are now 1d along path"""
    cdef public int pos_along_path
    cdef public np.float32_t[:,::1] path_xy
    cdef public np.float32_t[:,::1] path_ij # redundant
    cdef public CVectorFeatures path_features
    def __cinit__(self):
        pass
    # from path state
    def from_state_and_path(self, state, path_xy, path_ij):
        self.cfrom_state_and_path(state, path_xy, path_ij)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    @cython.cdivision(True)
    cdef cfrom_state_and_path(self, CState state, np.float32_t[:,::1] path_xy, np.float32_t[:,::1] path_ij):
        self.pos_along_path = index_along_path_from_pos_ij(state.pos_ij, path_ij)
        self.path_xy = path_xy
        self.path_ij = path_ij
        self.path_features = CVectorFeatures(path_ij.shape[0])
        cdef int n
        cdef int k
        cdef int i
        cdef int j
        for n in range(N_STATE_FEATURES):
            for k in range(path_ij.shape[0]):
                i = int(path_ij[k][0])
                j = int(path_ij[k][1])
                self.path_features._values()[n, k] = state.grid_features._values()[n, i, j]
                self.path_features._uncertainties()[n, k] = state.grid_features._uncertainties()[n, i, j]
    def copy(self):
        return self.ccopy()
    cdef CPathState ccopy(self):
        copy_ = CPathState()
        copy_.pos_along_path = self.pos_along_path
        copy_.path_xy = self.path_xy
        copy_.path_ij = self.path_ij
        copy_.path_features = CVectorFeatures(self.path_ij.shape[0])
        copy_.path_features.copy_from(self.path_features)
        return copy_
    def path_features_values(self):
        return np.array(self.path_features._values())
    def path_features_uncertainties(self):
        return np.array(self.path_features._uncertainties())
    def get_path_len(self):
        return len(self.path_xy)
    def get_path_xy(self):
        return np.array(self.path_xy)
    def get_path_ij(self):
        return np.array(self.path_ij)
def path_state_from_state_and_path(state, path_xy, path_ij):
    path_state = CPathState()
    path_state.from_state_and_path(state, path_xy, path_ij)
    return path_state

cdef class CTaskState:
    """ Reduction of State: 2d elements are now 1d, for single action along a subpath"""
    cdef int startidx_in_fullpath
    cdef np.float32_t[:,::1] subpath_xy
    cdef np.float32_t[:,::1] subpath_ij
    cdef CVectorFeatures task_features
    cdef np.float32_t resolution
    # from path state
    def __cinit__(self, CPathState path_state, int startidx, int endidx, np.float32_t resolution):
        self.cfrom_path_state(path_state, startidx, endidx, resolution)
    cdef cfrom_path_state(self, CPathState path_state, int startidx, int endidx, np.float32_t resolution):
        self.task_features = CVectorFeatures(endidx+1 - startidx)
        self.subpath_xy = path_state.path_xy[startidx:endidx+1,:]
        self.subpath_ij = path_state.path_ij[startidx:endidx+1,:]
        self.resolution = resolution
        self.startidx_in_fullpath = startidx
        # along subpath
        self.task_features._values()[:,:] = path_state.path_features._values()[:,startidx:endidx+1]
        self.task_features._uncertainties()[:,:] = path_state.path_features._uncertainties()[:,startidx:endidx+1]
    def get_subpath_xy(self):
        return np.array(self.subpath_xy)
    def get_subpath_ij(self):
        return np.array(self.subpath_ij)
    def task_features_values(self):
        return self.task_features._values()
    def task_features_uncertainties(self):
        return self.task_features._uncertainties()
    def path_length(self):
        return self.cpath_length()
    cdef np.float32_t cpath_length(self):
        cdef np.float32_t length = len(self.subpath_ij) * self.resolution
        return length
cdef class CTaskStateUpdate:
    """ Diff for a TaskState, obtained from predict_outcome """
    cdef bool is_pos_update
    cdef np.float32_t[:] pos_update
    cdef np.float32_t[:] pos_ij_update
    # task definition
    cdef int startidx_in_fullpath
    cdef np.float32_t[:,::1] subpath_xy
    cdef np.float32_t[:,::1] subpath_ij
    cdef CVectorFeatures task_features
    cdef CFeaturesDiff features_diff
    def __cinit__(
                self, CTaskState task_state,
                bool is_pos_update,
                np.float32_t[:] pos_update,
                np.float32_t[:] pos_ij_update,
                str action_name,
                str success,
            ):
        self.cgenerate_update_properties(
            task_state,
            is_pos_update,
            pos_update,
            pos_ij_update,
            action_name,
            success,
        )
    cdef cgenerate_update_properties(
                self, CTaskState task_state,
                bool is_pos_update,
                np.float32_t[:] pos_update,
                np.float32_t[:] pos_ij_update,
                str action_name,
                str success,
            ):
        self.is_pos_update = is_pos_update
        if is_pos_update:
            self.pos_update = pos_update
            self.pos_ij_update = pos_ij_update
        else:
            self.pos_update = np.zeros((2,), dtype=np.float32)
            self.pos_ij_update = np.zeros((2,), dtype=np.float32)
        # task definition
        self.startidx_in_fullpath = task_state.startidx_in_fullpath
        self.subpath_xy = task_state.subpath_xy
        self.subpath_ij = task_state.subpath_ij
        self.task_features = task_state.task_features
        # diff
        self.features_diff = CFeaturesDiff()
        # Fill in diffs
        if action_name == "INTEND":
            if success == "S":
                self.features_diff.is_changed[PERMISSIVITY_INDEX] = True
                self.features_diff.value_operations[PERMISSIVITY_INDEX] = INTEND_S_PERMISSIVITY_value_operation
                self.features_diff.uncertainty_operations[PERMISSIVITY_INDEX] = INTEND_S_PERMISSIVITY_uncertainty_operation
            else:
                self.features_diff.is_changed[PERMISSIVITY_INDEX] = True
                self.features_diff.value_operations[PERMISSIVITY_INDEX] = INTEND_F_PERMISSIVITY_value_operation
                self.features_diff.uncertainty_operations[PERMISSIVITY_INDEX] = INTEND_F_PERMISSIVITY_uncertainty_operation
                self.features_diff.is_changed[PERCEPTIVITY_INDEX] = True
                self.features_diff.value_operations[PERCEPTIVITY_INDEX] = INTEND_F_PERCEPTIVITY_value_operation
                self.features_diff.uncertainty_operations[PERCEPTIVITY_INDEX] = INTEND_F_PERCEPTIVITY_uncertainty_operation
        if action_name == "SAY":
            if success == "S":
                self.features_diff.is_changed[PERMISSIVITY_INDEX] = True
                self.features_diff.value_operations[PERMISSIVITY_INDEX] = SAY_S_PERMISSIVITY_value_operation
                self.features_diff.uncertainty_operations[PERMISSIVITY_INDEX] = SAY_S_PERMISSIVITY_uncertainty_operation
            else:
                self.features_diff.is_changed[PERMISSIVITY_INDEX] = True
                self.features_diff.value_operations[PERMISSIVITY_INDEX] = SAY_F_PERMISSIVITY_value_operation
                self.features_diff.uncertainty_operations[PERMISSIVITY_INDEX] = SAY_F_PERMISSIVITY_uncertainty_operation
                self.features_diff.is_changed[PERCEPTIVITY_INDEX] = True
                self.features_diff.value_operations[PERCEPTIVITY_INDEX] = SAY_F_PERCEPTIVITY_value_operation
                self.features_diff.uncertainty_operations[PERCEPTIVITY_INDEX] = SAY_F_PERCEPTIVITY_uncertainty_operation
        if action_name == "CRAWL":
            if success == "S":
                pass
            else:
                self.features_diff.is_changed[PERMISSIVITY_INDEX] = True
                self.features_diff.value_operations[PERMISSIVITY_INDEX] = CRAWL_F_PERMISSIVITY_value_operation
                self.features_diff.uncertainty_operations[PERMISSIVITY_INDEX] = CRAWL_F_PERMISSIVITY_uncertainty_operation
        if action_name == "NUDGE":
            if success == "S":
                pass
            else:
                self.features_diff.is_changed[PERMISSIVITY_INDEX] = True
                self.features_diff.value_operations[PERMISSIVITY_INDEX] = NUDGE_F_PERMISSIVITY_value_operation
                self.features_diff.uncertainty_operations[PERMISSIVITY_INDEX] = NUDGE_F_PERMISSIVITY_uncertainty_operation
    def apply_to_state_along_taskpath(self, CState state):
        return self.capply_to_state_along_taskpath(state)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    @cython.cdivision(True)
    cdef bool capply_to_state_along_taskpath(self, CState state):
        """ Applies diff along task path to state, inplace. UNUSED (State remains constant) """
        cdef bool is_any_changed = False
        cdef int n
        cdef int k
        cdef int i
        cdef int j
        for n in range(N_STATE_FEATURES):
            if self.features_diff.is_changed[n]:
                is_any_changed = True
                break
        if is_any_changed:
            # TODO currently diff is point_to_point, change? (circular stencils, maybe?)
            for k in range(self.subpath_ij.shape[0]):
                i = int(self.subpath_ij[k, 0])
                j = int(self.subpath_ij[k, 1])
                for n in range(N_STATE_FEATURES):
                    if self.features_diff.is_changed[n]:
                        state.grid_features.values[n, i, j] = self.features_diff.value_operations[n](
                            state.grid_features.values[:, i, j],
                            state.grid_features.uncertainties[:, i, j]
                        )
                        state.grid_features.uncertainties[n, i, j] = self.features_diff.uncertainty_operations[n](
                            state.grid_features.values[:, i, j],
                            state.grid_features.uncertainties[:, i, j]
                        )
        # update the position if it has changed
        if self.is_pos_update:
            is_any_changed = True
            state.pos = self.pos_update
            state.pos_ij = self.pos_ij_update
        return is_any_changed
    def apply_to_path_state(self, CPathState path_state):
        return self.capply_to_path_state(path_state)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    @cython.cdivision(True)
    cdef bool capply_to_path_state(self, CPathState path_state):
        """ Applies diff along task path to path state, inplace """
        cdef bool is_any_changed = False
        cdef int n
        cdef int k
        cdef int i
        cdef int j
        cdef int k_start = self.startidx_in_fullpath
        cdef int k_end = self.startidx_in_fullpath + self.subpath_ij.shape[0]
        for n in range(N_STATE_FEATURES):
            if self.features_diff.is_changed[n]:
                is_any_changed = True
                break
        if is_any_changed:
            # TODO currently diff is point_to_point, change? (circular stencils, maybe?)
            for k in range(k_start, k_end):
                for n in range(N_STATE_FEATURES):
                    if self.features_diff.is_changed[n]:
                        path_state.path_features.values[n, k] = \
                            self.features_diff.value_operations[n](
                                path_state.path_features.values[:, k],
                                path_state.path_features.uncertainties[:, k]
                            )
                        path_state.path_features.uncertainties[n, k] = \
                            self.features_diff.uncertainty_operations[n](
                                path_state.path_features.values[:, k],
                                path_state.path_features.uncertainties[:, k]
                            )
        # update the position if it has changed
        if self.is_pos_update:
            is_any_changed = True
            path_state.pos_along_path = k_end-1
        return is_any_changed

cdef class CState:
    """ storage class for convenience 
    Simplified state used by IA from single agent POV over the whole map """
    cdef float radius
    cdef np.float32_t[:] pos
    cdef np.float32_t[:] pos_ij # not rounded. redundant, for performance
    cdef np.float32_t[:] goal
    cdef CGridFeatures grid_features
    def __cinit__(self, radius, pos, pos_ij, mapwidth, mapheight, goal):
        # would like to do away with these
        self.radius = radius
        self.pos = pos
        self.pos_ij = pos_ij
        self.goal = goal
        # mathematical state features
        self.grid_features = CGridFeatures(mapwidth, mapheight)
    def copy(self):
        return self.ccopy()
    cdef CState ccopy(self):
        copy_ = CState(self.radius, self.pos, self.pos_ij,
                       self.grid_features.grid_width, self.grid_features.grid_height, self.goal)
        copy_.grid_features.copy_from(self.grid_features)
        return copy_
    def grid_features_values(self):
        return np.asarray(self.grid_features._values())
    def grid_features_uncertainties(self):
        return np.asarray(self.grid_features._uncertainties())
    def get_pos_ij(self):
        return np.asarray(self.pos_ij)
    def get_pos(self):
        return np.asarray(self.pos)
    def get_goal(self):
        return np.asarray(self.goal)
    def get_radius(self):
        return self.radius
    def serialize(self):
        return {
            "radius": float(self.radius),
            "pos": np.array(self.pos),
            "pos_ij": np.array(self.pos_ij),
            "goal": np.array(self.goal),
            "grid_features": self.grid_features.serialize(),
        }
def unserialize_cstate(dict_):
    gf = unserialize_cgridfeatures(dict_["grid_features"])
    s = CState(
        dict_["radius"], dict_["pos"], dict_["pos_ij"], 
        dict_["grid_features"]["grid_width"], dict_["grid_features"]["grid_height"], dict_["goal"]
    )
    s.grid_features = gf
    return s



class FixedState(object):
    """ Stores the constant parts of the state separately, to reduce computation """
    def __init__(self, map_):
        self.map = map_
        self.derived_state = DerivedState()
    def serialize(self):
        return {
            "map": self.map.serialize(),
            "derived_state": None, # TODO actually serialize
        }
def unserialize_fixedstate(dict_):
    map_ = CMap2D()
    map_.unserialize(dict_["map"])
    fs = FixedState(map_)
    return fs


class DerivedState(object):
    """ convenience storage class for precomputed/cached fields,
    this class is for algorithmic optimization and thus isn't formally in the IA algo"""
    def __init__(self):
        self._map_sdf = None
        self._cache_map = None

        self._suppress_warnings = False

    # CORE INTERFACE -----------
    def copy(self):
        copy_ = DerivedState()
        copy_._map_sdf = self._map_sdf
        copy_._cache_map = self._cache_map
        return copy_

    def map_sdf(self, fixed_state):
        if fixed_state.map is not self._cache_map:
            self._on_set_map(fixed_state)
        return self._map_sdf

    def dijkstra_from_goal(self, state, fixed_state, path_variant):
        goal_ij = fixed_state.map.xy_to_ij([state.get_goal()])[0]
        sdf = self.map_sdf(fixed_state)
        is_traversable = path_variant.is_state_traversable(state, fixed_state)
        mask = np.logical_not(is_traversable).astype(np.uint8)
        extra_costs = (0.2/(0.00001+sdf**2) + 10 * (sdf < state.get_radius())).astype(np.float32)
        speeds = 1./(1.+extra_costs)
        return fixed_state.map.fastmarch(goal_ij, mask=mask, speeds=speeds)

    def path_to_goal(self, state, fixed_state, path_variant):
        agent_ij = state.get_pos_ij()
        dijkstra = self.dijkstra_from_goal(state, fixed_state, path_variant)
        path_ij, _ = path_from_dijkstra_field(dijkstra, agent_ij, connectedness=8)
        path_xy = fixed_state.map.ij_to_xy(path_ij)
        # check if path reaches goal
        goal_ij = fixed_state.map.xy_to_ij([state.get_goal()])[0]
        if not np.allclose(path_ij[-1], goal_ij):
#             print("Path does not reach goal.")
            return None
        return path_xy

    def suppress_warnings(self, set_to=True):
        self._suppress_warnings = set_to


    # INTERNAL --------------
    def _on_set_map(self, fixed_state):
        if not self._suppress_warnings:
            print("DerivedState: Precomputing cached sdf")
        self._map_sdf = fixed_state.map.as_sdf()
        self._cache_map = fixed_state.map


cdef int index_along_path_from_pos_ij(np.float32_t[:] pos_ij, np.float32_t[:,::1] path_ij):
    cdef int n
    cdef np.float32_t i
    cdef np.float32_t j
    for n in range(path_ij.shape[0]):
        i = path_ij[n,0]
        j = path_ij[n,1]
        if cfloor(i) == cfloor(pos_ij[0]) and cfloor(j) == cfloor(pos_ij[1]):
            return n
    raise ValueError("State position {} not in path {}".format(pos_ij, path_ij))

# Utilitites
# ----------------------------------
from libc.stdlib cimport rand, RAND_MAX
def sultans_wife(probs):
    """
    see http://datagenetics.com/blog/november52019/index.html
    assumes probs sum up to 1
    """
    return csultans_wife(probs)
cdef int csultans_wife(np.float32_t[:] probs):
    cdef np.float32_t r
    cdef int i
    cdef np.float32_t remaining_p_space = 1.
    cdef np.float32_t prob
    cdef np.float32_t rescaled_prob # probability rescaled to remaining p space
    for i in range(len(probs)):
        prob = probs[i]
        rescaled_prob = prob / remaining_p_space
        r = np.float32(rand())/np.float32(RAND_MAX) # random integer 0-1
        if r <= rescaled_prob:
            return i
        else:
            remaining_p_space -= prob
    raise ValueError

def segment_endindex_at_pathpos_i(np.int16_t[:,::1] segments, CPathState path_state):
    """ 
    segments              endindices
    [[0, 2], [4, 6]] --> [2,2,2,-1,6,6,6]
    """
    # -1 represents an invalid value, not the end of the path!
    endindices = -1 * np.ones((path_state.path_xy.shape[0],), dtype=np.int16)
    csegment_endindex_at_pathpos_i(segments, endindices, path_state)
    return endindices
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef void csegment_endindex_at_pathpos_i(np.int16_t[:,::1] segments, np.int16_t[:] endindices, CPathState path_state):
    cdef np.int16_t n
    cdef np.int16_t i
    for n in range(segments.shape[0]):
        for i in range(segments[n,0], segments[n,1]+1):
            endindices[i] = segments[n,1]

def segments_from_bool1d(bool1d):
    """
    transform list of bools to segments where bool1d is true
    11100011111110 -> 2 segments, {(0,2), (6, 12)} """
    segments = np.zeros((cn_segments_from_bool1d(bool1d), 2), dtype=np.int16)
    csegments_from_bool1d(bool1d, segments)
    return segments
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef int cn_segments_from_bool1d(np.uint8_t[:] bool1d):
    cdef int i
    cdef int n_segments = 0
    cdef np.uint8_t segment_is_open = False
    for i in range(bool1d.shape[0]):
        if not segment_is_open:
            if bool1d[i]:
                segment_is_open = True
            continue
        else:
            if bool1d[i]:
                continue
            else:
                n_segments += 1
                segment_is_open = False
    if segment_is_open:
        n_segments += 1
        segment_is_open = False
    return n_segments
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef csegments_from_bool1d(np.uint8_t[:] bool1d, np.int16_t[:,::1] segments):
    cdef int i
    cdef int segment_index = 0
    cdef np.uint8_t segment_is_open = False
    for i in range(bool1d.shape[0]):
        if not segment_is_open:
            if bool1d[i]:
                # open new segment
                segment_is_open = True
                segments[segment_index, 0] = i
            continue
        else:
            if bool1d[i]:
                continue
            else:
                segments[segment_index, 1] = i
                segment_is_open = False
                segment_index += 1
    if segment_is_open:
        segments[segment_index, 1] = i


def bool1d_from_segments(segments, length=None):
    """
    reverse of segments_from_bool1d operation
    11100011111110 <- 2 segments, {(0,2), (6, 12)} """
    if length is None:
        length = int(0)
        clength_bool1d_from_segments(segments, length)
    bool1d = np.zeros((length), dtype=np.uint8)
    cbool1d_from_segments(segments, bool1d)
    return bool1d

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef void clength_bool1d_from_segments(np.int16_t[:,::1] segments, int length):
    cdef np.int16_t end = 0
    cdef int i
    for i in range(segments.shape[0]):
        end = segments[i,1]
        if end > length:
            length = end
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef void cbool1d_from_segments(np.int16_t[:,::1] segments, np.uint8_t[:] bool1d):
    cdef int i
    cdef int n
    for n in range(segments.shape[0]):
        for i in range(segments[n,0], segments[n,1]+1):
            bool1d[i] = True

@cython.nonecheck(False)
cdef np.float32_t cclip(np.float32_t val, np.float32_t min_, np.float32_t max_):
    cdef np.float32_t result = val
    if val < min_:
        result = min_
    if val > max_:
        result = max_
    return result

# Below are success probabilities for actions P(Success | S)
# ---------------------------------------------------------------------------------
def INTEND_predict_success(task_state):
    return cINTEND_predict_success(task_state)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef np.float32_t cINTEND_predict_success(CTaskState task_state):
    # probability as a function of state
    cdef int k
    cdef int k_len = len(task_state.subpath_ij)
    cdef np.float32_t k_length = k_len * task_state.resolution
    cdef np.float32_t path_prob
    cdef np.float32_t cumulative_prob = 1.
    cdef np.float32_t permissivity_factor
    cdef np.float32_t perceptivity_factor
    cdef np.float32_t crowdedness_factor
    cdef np.float32_t local_prob
    cdef np.float32_t distance_equivalent_prob
    # sample values from distribution (take max prob i.e. mean)
    cdef np.float32_t[:,::1] values = task_state.task_features_values()
    cdef np.float32_t[:,::1] uncertainties = task_state.task_features_uncertainties()
    for k in range(k_len):
        permissivity_factor = values[PERMISSIVITY_INDEX, k]
        perceptivity_factor = cclip(values[PERCEPTIVITY_INDEX, k], 0., 1.)
        crowdedness_factor = 0.5 / cclip(values[CROWDEDNESS_INDEX, k], 1., 10.)
        # probability of success for single step in path ( if that step is 10m long )
        local_prob = permissivity_factor * perceptivity_factor * crowdedness_factor
        cumulative_prob = min(cumulative_prob, local_prob)  # the right way (geometric mean) is numerically unstable
    path_prob = cclip(cumulative_prob, 0.05, 0.99)
    # convert from reference step size (10m) to actual size
    distance_equivalent_prob = path_prob ** (k_length / PROB_DISTANCE_REF)
    # shortcut (numerically unstable)
#     cumulative_prob = cumulative_prob * local_prob
#     distance_equivalent_prob = cumulative_prob ** (task_state.resolution / PROB_DISTANCE_REF)
    return distance_equivalent_prob

def SAY_predict_success(task_state):
    return cSAY_predict_success(task_state)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef np.float32_t cSAY_predict_success(CTaskState task_state):
    # probability as a function of state
    cdef int k
    cdef int k_len = len(task_state.subpath_ij)
    cdef np.float32_t k_length = k_len * task_state.resolution
    cdef np.float32_t path_prob
    cdef np.float32_t cumulative_prob = 1.
    cdef np.float32_t permissivity_bonus = 1.5
    cdef np.float32_t permissivity_factor
    cdef np.float32_t perceptivity_factor = 1.  # saying action should have maxed PER
    cdef np.float32_t crowdedness_factor
    cdef np.float32_t local_prob
    cdef np.float32_t distance_equivalent_prob
    # sample values from distribution (take max prob i.e. mean)
    cdef np.float32_t[:,::1] values = task_state.task_features_values()
    cdef np.float32_t[:,::1] uncertainties = task_state.task_features_uncertainties()
    for k in range(k_len):
        permissivity_factor = permissivity_bonus * values[PERMISSIVITY_INDEX, k]
        crowdedness_factor = 1. / cclip(values[CROWDEDNESS_INDEX, k], 1., 10.)
        # probability of success for single step in path ( if that step is 10m long )
        local_prob = permissivity_factor * perceptivity_factor * crowdedness_factor
        cumulative_prob = min(cumulative_prob, local_prob)
    path_prob = cclip(cumulative_prob, 0.05, 0.99)
    # convert from reference step size (10m) to actual size
    distance_equivalent_prob = path_prob ** (k_length / PROB_DISTANCE_REF)
    # shortcut (numerically unstable)
#     cumulative_prob = cumulative_prob * local_prob
#     distance_equivalent_prob = cumulative_prob ** (task_state.resolution / PROB_DISTANCE_REF)
    return distance_equivalent_prob

def NUDGE_predict_success(task_state):
    return cNUDGE_predict_success(task_state)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef np.float32_t cNUDGE_predict_success(CTaskState task_state):
    # probability as a function of state
    cdef int k
    cdef int k_len = len(task_state.subpath_ij)
    cdef np.float32_t k_length = k_len * task_state.resolution
    cdef np.float32_t path_prob
    cdef np.float32_t cumulative_prob = 1.
    cdef np.float32_t permissivity_bonus = 5.
    cdef np.float32_t permissivity_factor
    cdef np.float32_t perceptivity_factor = 1.  # perceptivity is not a problem
    cdef np.float32_t crowdedness_factor = 1.  # crowdedness is not a problem
    cdef np.float32_t local_prob
    cdef np.float32_t distance_equivalent_prob
    # sample values from distribution (take max prob i.e. mean)
    cdef np.float32_t[:,::1] values = task_state.task_features_values()
    cdef np.float32_t[:,::1] uncertainties = task_state.task_features_uncertainties()
    for k in range(k_len):
        permissivity_factor = permissivity_bonus * values[PERMISSIVITY_INDEX, k]
        # probability of success for single step in path ( if that step is 10m long )
        local_prob = permissivity_factor * perceptivity_factor * crowdedness_factor
        cumulative_prob = min(cumulative_prob, local_prob)
    path_prob = cclip(cumulative_prob, 0.05, 0.99)
    # convert from reference step size (10m) to actual size
    distance_equivalent_prob = path_prob ** (k_length / PROB_DISTANCE_REF)
    # shortcut (numerically unstable)
#     cumulative_prob = cumulative_prob * local_prob
#     distance_equivalent_prob = cumulative_prob ** (task_state.resolution / PROB_DISTANCE_REF)
    return distance_equivalent_prob

# Below are value operations  P(S' | Outcome) = T_outcome(S, b, Eps)
# ---------------------------------------------------------------------------------
#
# INTEND - SUCCESS - PERMISSIVITY
cdef np.float32_t INTEND_S_PERMISSIVITY_value_operation(np.float32_t[:] values, np.float32_t[:] uncertainties):
    if values[PERMISSIVITY_INDEX] > 0:
        return 1 - (1 - values[PERMISSIVITY_INDEX]) * 0.9 # if permissivity was low, increase it
    else:
        return 0.1
cdef np.float32_t INTEND_S_PERMISSIVITY_uncertainty_operation(np.float32_t[:] values, np.float32_t[:] uncertainties):
    return uncertainties[PERMISSIVITY_INDEX] * 0.1
# INTEND - FAILURE - PERMISSIVITY
cdef np.float32_t INTEND_F_PERMISSIVITY_value_operation(np.float32_t[:] values, np.float32_t[:] uncertainties):
    return values[PERMISSIVITY_INDEX] * 0.1
cdef np.float32_t INTEND_F_PERMISSIVITY_uncertainty_operation(np.float32_t[:] values, np.float32_t[:] uncertainties):
    return uncertainties[PERMISSIVITY_INDEX] * 0.001
# INTEND - FAILURE - PERCEPTIVITY
cdef np.float32_t INTEND_F_PERCEPTIVITY_value_operation(np.float32_t[:] values, np.float32_t[:] uncertainties):
    return values[PERCEPTIVITY_INDEX] * 0.5
cdef np.float32_t INTEND_F_PERCEPTIVITY_uncertainty_operation(np.float32_t[:] values, np.float32_t[:] uncertainties):
    return uncertainties[PERCEPTIVITY_INDEX] * 0.1

# SAY - SUCCESS - PERMISSIVITY
cdef np.float32_t SAY_S_PERMISSIVITY_value_operation(np.float32_t[:] values, np.float32_t[:] uncertainties):
    if values[PERMISSIVITY_INDEX] > 0:
        return 1 - (1 - values[PERMISSIVITY_INDEX]) * 0.9 # increase to 1
    else:
        return 0.1
cdef np.float32_t SAY_S_PERMISSIVITY_uncertainty_operation(np.float32_t[:] values, np.float32_t[:] uncertainties):
    return uncertainties[PERMISSIVITY_INDEX] * 0.1
# SAY - FAILURE - PERMISSIVITY
cdef np.float32_t SAY_F_PERMISSIVITY_value_operation(np.float32_t[:] values, np.float32_t[:] uncertainties):
    return values[PERMISSIVITY_INDEX] * 0.1  # decrease to 0
cdef np.float32_t SAY_F_PERMISSIVITY_uncertainty_operation(np.float32_t[:] values, np.float32_t[:] uncertainties):
    return uncertainties[PERMISSIVITY_INDEX] * 0.001
# SAY - FAILURE - PERCEPTIVITY
cdef np.float32_t SAY_F_PERCEPTIVITY_value_operation(np.float32_t[:] values, np.float32_t[:] uncertainties):
    return 1.
cdef np.float32_t SAY_F_PERCEPTIVITY_uncertainty_operation(np.float32_t[:] values, np.float32_t[:] uncertainties):
    return uncertainties[PERCEPTIVITY_INDEX] * 0.1

# CRAWL - FAILURE - PERMISSIVITY
cdef np.float32_t CRAWL_F_PERMISSIVITY_value_operation(np.float32_t[:] values, np.float32_t[:] uncertainties):
    return values[PERMISSIVITY_INDEX] * 0.25 # decrease to 0
cdef np.float32_t CRAWL_F_PERMISSIVITY_uncertainty_operation(np.float32_t[:] values, np.float32_t[:] uncertainties):
    return uncertainties[PERMISSIVITY_INDEX] * 0.1

# NUDGE - FAILURE - PERMISSIVITY
cdef np.float32_t NUDGE_F_PERMISSIVITY_value_operation(np.float32_t[:] values, np.float32_t[:] uncertainties):
    # maybe not? the nudge planner could simply be bad in high permissivity cases
    return values[PERMISSIVITY_INDEX] * 0.1 # decrease to 0
cdef np.float32_t NUDGE_F_PERMISSIVITY_uncertainty_operation(np.float32_t[:] values, np.float32_t[:] uncertainties):
    return uncertainties[PERMISSIVITY_INDEX] * 0.001


def apply_value_operations_to_state(update, state, update_stencil):
    return capply_value_operations_to_state(update, state, update_stencil)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef bool capply_value_operations_to_state(CTaskStateUpdate update, CState state,
                                          np.float32_t[:,::1] update_stencil):
    """ Applies diff to entire state (full grids), inplace - expensive """
    cdef bool is_any_changed = False
    cdef int n
    cdef int i
    cdef int j
    cdef int len_i = state.grid_features.grid_width
    cdef int len_j = state.grid_features.grid_height
    cdef np.float32_t update_strenth
    cdef np.float32_t updated_value
    cdef np.float32_t updated_uncertainty
    cdef np.float32_t old_value
    cdef np.float32_t old_uncertainty
    for n in range(N_STATE_FEATURES):
        if update.features_diff.is_changed[n]:
            is_any_changed = True
            break
    if is_any_changed:
        for i in range(len_i):
            for j in range(len_j):
                update_strength = update_stencil[i, j]
                for n in range(N_STATE_FEATURES):
                    if update.features_diff.is_changed[n]:
                        old_value = state.grid_features.values[n, i, j]
                        old_uncertainty = state.grid_features.uncertainties[n, i, j]
                        updated_value = update.features_diff.value_operations[n](
                            state.grid_features.values[:, i, j],
                            state.grid_features.uncertainties[:, i, j]
                        )
                        updated_uncertainty = update.features_diff.uncertainty_operations[n](
                            state.grid_features.values[:, i, j],
                            state.grid_features.uncertainties[:, i, j]
                        )
                        # apply
                        state.grid_features.values[n, i, j] = (
                            update_strength * updated_value + (1 - update_strength) * old_value
                        )
                        state.grid_features.uncertainties[n, i, j] = (
                            update_strength * updated_uncertainty + (1 - update_strength) * old_uncertainty
                        )
    # update the position if it has changed
    if update.is_pos_update:
        is_any_changed = True
        state.pos = update.pos_update
        state.pos_ij = update.pos_ij_update
    return is_any_changed



import numpy as np

from pyniel.numpy_tools import indexing

from cia import kStateFeatures as kSF

WARMUP_INITIAL_RADIUS = 1. # warmup means the path conditions are relaxed for the first few meters
class Path(object):
    def type(self):
        return type(self)

    def typestr(self):
        return str(self.type()).replace("<class 'pyIA.paths.", '').replace("'>", '')

    def letterstr(self):
        return self.typestr()[0]

    def linestyle(self):
        return '-'
    def linecolor(self):
        return 'black'

class NaivePath(Path):
    def linestyle(self):
        return '-'
    def linecolor(self):
        return 'tab:blue'
    def is_state_traversable(self, state, fixed_state):
        is_traversable = \
        fixed_state.map.occupancy() < fixed_state.map.thresh_occupied()
        return is_traversable
class PermissivePath(Path):
    def linestyle(self):
        return '-.'
    def linecolor(self):
        return 'orange'
    def is_state_traversable(self, state, fixed_state):
        is_traversable = \
        np.logical_and.reduce([
            np.logical_or( # high or unknown permissivity
                state.grid_features_values()[kSF["permissivity"]] >= 0.2, # high permissivity
                state.grid_features_uncertainties()[kSF["permissivity"]] >= 1., # low confidence
            ),
            np.logical_or( # low or unknown crowdedness
                state.grid_features_values()[kSF["crowdedness"]] <= 1., # low crowdedness
                state.grid_features_uncertainties()[kSF["crowdedness"]] >= 1., # low crowdedness
            ),
            fixed_state.map.occupancy() < fixed_state.map.thresh_occupied()
        ])
        return is_traversable
class StrictPermissivePath(Path):
    def linestyle(self):
        return '--'
    def linecolor(self):
        return 'tab:red'
    def is_state_traversable(self, state, fixed_state):
        is_traversable = \
        np.logical_and.reduce([
            np.logical_or( # very high or unknown permissivity
                state.grid_features_values()[kSF["permissivity"]] >= 0.8, # high permissivity
                state.grid_features_uncertainties()[kSF["permissivity"]] >= 1., # low confidence
            ),
            np.logical_or( # low or unknown crowdedness
                state.grid_features_values()[kSF["crowdedness"]] <= 1., # low crowdedness
                state.grid_features_uncertainties()[kSF["crowdedness"]] >= 1., # low crowdedness
            ),
            fixed_state.map.occupancy() < fixed_state.map.thresh_occupied()
        ])
        return is_traversable
class WarmupStrictPermissivePath(Path):
    def linestyle(self):
        return ':'
    def linecolor(self):
        return 'tab:purple'
    def is_state_traversable(self, state, fixed_state):
        sqr_initial_radius_ij = (WARMUP_INITIAL_RADIUS / fixed_state.map.resolution())**2
        ij = indexing.as_idx_array(fixed_state.map.occupancy(), axis='all')
        ii = ij[...,0]
        jj = ij[...,1]
        ai, aj = fixed_state.map.xy_to_floatij([state.get_pos()])[0][:2]
        is_in_initial_radius = ((ii - ai)**2 + (jj - aj)**2) < sqr_initial_radius_ij
        is_traversable = \
        np.logical_and(
            np.logical_or(
                np.logical_and(
                    np.logical_or( # very high or unknown permissivity
                        state.grid_features_values()[kSF["permissivity"]] >= 0.8, # high permissivity
                        state.grid_features_uncertainties()[kSF["permissivity"]] >= 1., # low confidence
                    ),
                    np.logical_or( # low or unknown crowdedness
                        state.grid_features_values()[kSF["crowdedness"]] <= 1., # low crowdedness
                        state.grid_features_uncertainties()[kSF["crowdedness"]] >= 1., # low crowdedness
                    )
                ),
                is_in_initial_radius
            ),
            fixed_state.map.occupancy() < fixed_state.map.thresh_occupied()
        )
        return is_traversable

PATH_VARIANTS = [NaivePath(), PermissivePath(), StrictPermissivePath(), WarmupStrictPermissivePath()]

def is_path_equivalent(a_path, b_path):
    if not np.allclose(np.shape(a_path), np.shape(b_path)):
        return False
    return np.allclose(a_path, b_path)

def is_path_in_list(path, list_):
    for b_path in list_:
        if is_path_equivalent(path, b_path):
            return True
    return False
